import contextlib, flask, functools, pathlib, traceback, urllib
import http.client
from . import server
from .. import control, editor
from ... util import data_file, log
from ... util.threads import runnable

PORT = 8787
ROOT_FOLDER = pathlib.Path(__file__).parents[3] / 'ui' / 'rest'
INDEX_FILE = 'index.html'
BAD_ADDRESS_ERROR = 'Bad address {address}'
BAD_GETTER_ERROR = 'Couldn\'t get address {address}'
BAD_SETTER_ERROR = 'Couldn\'t set value {value} at address {address}'


def _single(method):
    """Decorator for Rest methods that take a single address"""
    @functools.wraps(method)
    def single(self, address, value=None):
        address = urllib.parse.unquote_plus(address)
        try:
            error = BAD_ADDRESS_ERROR
            ed = editor.Editor(address, self.root)

            if value is None:
                error = BAD_GETTER_ERROR
                result = method(self, ed)
            else:
                error = BAD_SETTER_ERROR
                result = method(self, ed, value)
            result = {'value': result}

        except Exception as e:
            if self.verbose:
                traceback.print_exc()
            msg = '%s\n%s' % (error.format(**locals()), e)
            result = {'error': msg}

        return flask.jsonify(result)

    return single


def _multi(method):
    """Decorator for Rest methods that take multiple addresses"""
    @functools.wraps(method)
    def multi(self, address=''):
        values = flask.request.values
        address = urllib.parse.unquote_plus(address)
        if address and values and not address.endswith('.'):
            address += '.'

        result = {}
        for a in values or '':
            try:
                ed = editor.Editor(address + a, self.root)
                result[address + a] = {'value': method(self, ed, a)}
            except:
                if self.verbose:
                    traceback.print_exc()
                result[address + a] = {'error': 'Could not multi addr %s' % a}

        return flask.jsonify(result)

    return multi


class Rest(control.ExtractedControl):
    OPEN_PAGE = False

    def __init__(self, *args, port=PORT, external_access=False, verbose=True,
                 root_folder=ROOT_FOLDER, index_file=INDEX_FILE,
                 open_page=None, **kwds):
        super().__init__(*args, verbose=verbose, **kwds)

        root_folder = pathlib.Path(root_folder)
        index_file = pathlib.Path(index_file)
        if not index_file.is_absolute():
            index_file = root_folder / index_file
        self.index_file = str(index_file)

        static = str(root_folder / 'static')

        if open_page is None:
            open_page = self.OPEN_PAGE

        self.server = server.Server(
            port, external_access, open_page, static_folder=static)

        app = self.server.app
        app.route('/')(self.index)

        # Basic Endpoints
        app.route('/get/<address>')(self.get)
        app.route('/set/<address>/<value>')(self.put)
        # <address> and <value> are URL quoted

        # GET endpoints
        app.route('/single/<address>')(self.get)

        app.route('/multi')(self.multi_get)
        app.route('/multi/<address>')(self.multi_get)

        # PUT endpoints
        app.route('/single/<address>', methods=['PUT'])(self.single_put)

        app.route('/multi', methods=['PUT'])(self.multi_put)
        app.route('/multi/<address>', methods=['PUT'])(self.multi_put)

    def _make_thread(self):
        return self.server

    def index(self):
        with open(str(self.index_file)) as fp:
            return fp.read()

    @_single
    def get(self, editor):
        return self._get(editor)

    @_single
    def put(self, editor, value):
        unquoted = urllib.parse.unquote_plus(value)
        return self._set(editor, unquoted)

    @_single
    def single_put(self, editor):
        value = flask.request.values['value']
        return self._set(editor, value)

    @_multi
    def multi_get(self, editor, address):
        return self._get(editor)

    @_multi
    def multi_put(self, editor, address):
        value = flask.request.values[address]
        return self._set(editor, value)

    def _get(self, editor):
        return data_file.dumps(editor.get(), use_yaml=False)

    def _set(self, editor, value):
        loaded = data_file.loads(value)
        editor.set(loaded)
        return True

    @contextlib.contextmanager
    def _abort(self, code):
        try:
            yield
        except:
            if self.verbose:
                traceback.print_exc()
            flask.abort(code)


class OpenPage(Rest):
    OPEN_PAGE = True



from . import aliases
import loady, os
from .. util import data_file, log

guess_name = loady.importer.guess_name
CACHE = os.path.expanduser('~/.bibliopixel/code_cache')
ROOT_FILE = None


def data(name, use_json=True):
    if not name:
        return {}

    try:
        return data_file.loads(name)
    except:
        return loady.data.load(name, use_json)


def code(name, python_path=None):
    return name and loady.code.load_code(name, python_path, recurse=True)


def module(name, python_path=None):
    return name and loady.code.load_module(name, python_path)


def extender(path):
    parts = [os.getcwd()]
    if path:
        parts.extend(path.split(':'))

    missing = [p for p in parts if not os.path.exists(p)]
    if missing:
        msg = ('This "path" entry does not exist' if len(missing) == 1
               else 'These "path" entries do not exist')
        m2 = ['"%s"' % p for p in missing]
        log.warning('%s: %s', msg, ', '.join(m2))
        parts = [p for p in parts if os.path.exists(p)]

    return parts and loady.sys_path.extender(':'.join(parts))


def load_if_filename(s):
    if isinstance(s, str) and (s.endswith('.yml') or s.endswith('.json')):
        if ROOT_FILE and not os.path.isabs(s):
            s = os.path.join(os.path.dirname(ROOT_FILE), s)

        return data_file.load(s)



from .. import log, util
import colorsys


def hsv2rgb_raw(hsv):
    """
    Converts an HSV tuple to RGB. Intended for internal use.
    You should use hsv2rgb_spectrum or hsv2rgb_rainbow instead.
    """

    HSV_SECTION_3 = 0x40

    h, s, v = hsv

    # The brightness floor is minimum number that all of
    # R, G, and B will be set to.
    invsat = 255 - s
    brightness_floor = (v * invsat) // 256

    # The color amplitude is the maximum amount of R, G, and B
    # that will be added on top of the brightness_floor to
    # create the specific hue desired.
    color_amplitude = v - brightness_floor

    # figure out which section of the hue wheel we're in,
    # and how far offset we are within that section
    section = h // HSV_SECTION_3  # 0..2
    offset = h % HSV_SECTION_3  # 0..63

    rampup = offset
    rampdown = (HSV_SECTION_3 - 1) - offset

    # compute color-amplitude-scaled-down versions of rampup and rampdown
    rampup_amp_adj = (rampup * color_amplitude) // (256 // 4)
    rampdown_amp_adj = (rampdown * color_amplitude) // (256 // 4)

    # add brightness_floor offset to everything
    rampup_adj_with_floor = rampup_amp_adj + brightness_floor
    rampdown_adj_with_floor = rampdown_amp_adj + brightness_floor

    r, g, b = (0, 0, 0)

    if section:
        if section == 1:
            # section 1: 0x40..0x7F
            r = brightness_floor
            g = rampdown_adj_with_floor
            b = rampup_adj_with_floor
        else:
            # section 2; 0x80..0xBF
            r = rampup_adj_with_floor
            g = brightness_floor
            b = rampdown_adj_with_floor
    else:
        # section 0: 0x00..0x3F
        r = rampdown_adj_with_floor
        g = rampup_adj_with_floor
        b = brightness_floor

    return (r, g, b)


def hsv2rgb_spectrum(hsv):
    """Generates RGB values from HSV values in line with a typical light
    spectrum."""
    h, s, v = hsv
    return hsv2rgb_raw(((h * 192) >> 8, s, v))


def hsv2rgb_rainbow(hsv):
    """Generates RGB values from HSV that have an even visual
    distribution.  Be careful as this method is only have as fast as
    hsv2rgb_spectrum."""

    def nscale8x3_video(r, g, b, scale):
        nonzeroscale = 0
        if scale != 0:
            nonzeroscale = 1
        if r != 0:
            r = ((r * scale) >> 8) + nonzeroscale
        if g != 0:
            g = ((g * scale) >> 8) + nonzeroscale
        if b != 0:
            b = ((b * scale) >> 8) + nonzeroscale
        return (r, g, b)

    def scale8_video_LEAVING_R1_DIRTY(i, scale):
        nonzeroscale = 0
        if scale != 0:
            nonzeroscale = 1
        if i != 0:
            i = ((i * scale) >> 8) + nonzeroscale
        return i

    h, s, v = hsv
    offset = h & 0x1F  # 0..31
    offset8 = offset * 8
    third = (offset8 * (256 // 3)) >> 8
    r, g, b = (0, 0, 0)

    if not (h & 0x80):
        if not (h & 0x40):
            if not (h & 0x20):
                r = 255 - third
                g = third
                b = 0
            else:
                r = 171
                g = 85 + third
                b = 0x00
        else:
            if not (h & 0x20):
                twothirds = (third << 1)
                r = 171 - twothirds
                g = 171 + third
                b = 0
            else:
                r = 0
                g = 255 - third
                b = third
    else:
        if not (h & 0x40):
            if not (h & 0x20):
                r = 0x00
                twothirds = (third << 1)
                g = 171 - twothirds
                b = 85 + twothirds
            else:
                r = third
                g = 0
                b = 255 - third
        else:
            if not (h & 0x20):
                r = 85 + third
                g = 0
                b = 171 - third
            else:
                r = 171 + third
                g = 0x00
                b = 85 - third

    if s != 255:
        r, g, b = nscale8x3_video(r, g, b, s)
        desat = 255 - s
        desat = (desat * desat) >> 8
        brightness_floor = desat
        r = r + brightness_floor
        g = g + brightness_floor
        b = b + brightness_floor

    if v != 255:
        v = scale8_video_LEAVING_R1_DIRTY(v, v)
        r, g, b = nscale8x3_video(r, g, b, v)

    return (r, g, b)


def hsv2rgb_360(hsv):
    """Python default hsv to rgb conversion for when hue values in the
    range 0-359 are preferred.  Due to requiring float math, this method
    is slower than hsv2rgb_rainbow and hsv2rgb_spectrum."""

    h, s, v = hsv

    r, g, b = colorsys.hsv_to_rgb(h / 360.0, s, v)
    return (int(r * 255.0), int(g * 255.0), int(b * 255.0))


# pre-generated spectra for the sake of speed
HUE_RAW = [hsv2rgb_raw((hue, 255, 255)) for hue in range(256)]
HUE_RAINBOW = [hsv2rgb_rainbow((hue, 255, 255)) for hue in range(256)]
HUE_SPECTRUM = [hsv2rgb_spectrum((hue, 255, 255)) for hue in range(256)]
HUE_360 = [hsv2rgb_360((hue, 1.0, 1.0)) for hue in range(360)]


def hue2rgb_raw(hue):
    if hue >= 0 or hue < 256:
        return HUE_RAW[int(hue)]
    else:
        raise ValueError("hue must be between 0 and 255")


def hue2rgb_rainbow(hue):
    if hue >= 0 or hue < 256:
        return HUE_RAINBOW[int(hue)]
    else:
        raise ValueError("hue must be between 0 and 255")


def hue2rgb_spectrum(hue):
    if hue >= 0 or hue < 256:
        return HUE_SPECTRUM[int(hue)]
    else:
        raise ValueError("hue must be between 0 and 255")


def hue2rgb_360(hue):
    if hue >= 0 or hue < 360:
        return HUE_360[int(hue)]
    else:
        raise ValueError("hue must be between 0 and 359")


hsv2rgb = hsv2rgb_rainbow
hue2rgb = hue2rgb_rainbow


def hue_gradient(start, stop, steps):
    if not (0 <= start <= 255 and 0 <= stop <= 255):
        log.error(
            'hue must be between 0 and 255; start=%s, stop=%s', start, stop)
        start = min(255, max(0, start))
        stop = min(255, max(0, stop))
    flip = False
    if start > stop:
        start, stop = stop, start
        flip = True

    stops = util.even_dist(start, stop, steps)

    if flip:
        stops = stops[::-1]

    return stops


def hue_helper(pos, length, cycle_step):
    return hue2rgb(((pos * 255 // length) + cycle_step) % 255)


def hue_helper360(pos, length, cycle_step):
    return hue2rgb_360(((pos * 360 // length) + cycle_step) % 360)


def rgb_to_hsv(pixel):
    return colorsys.rgb_to_hsv(*(p / 255 for p in pixel))


def color_cmp(a, b):
    """Order colors by hue, saturation and value, in that order.

    Returns -1 if a < b, 0 if a == b and 1 if a < b.
    """
    if a == b:
        return 0

    a, b = rgb_to_hsv(a), rgb_to_hsv(b)
    return -1 if a < b else 1



from . import collection


class Parallel(collection.Collection):
    """
    A Parallel is a Collection where all the animations are running at the same
    time - as opposed to the base Collection where only one animation runs
    at any time.
    """

    def __init__(self, *args, overlay=False, detach=True, **kwds):
        """
        If overlay is True, then preclear is set to False for everything
        other than the first animation.
        """
        super().__init__(*args, **kwds)
        if detach:
            self.detach(overlay)

    def generate_frames(self):
        self._frames = [[a.generate_frames(), True] for a in self.animations]
        return super().generate_frames()

    def step(self, amt=1):
        for i, (frames, enabled) in enumerate(self._frames):
            if enabled:
                try:
                    next(frames)
                except StopIteration:
                    self._frames[i][1] = False



"""
A server that queues and sends UDP requests for a specific port on a
separate thread.
"""

import queue, socket
from .. util.threads import runnable
from .. util import log


class Sender(runnable.Runnable):
    def __init__(self, address):
        """
        :param str address: a pair (ip_address, port) to pass to socket.connect
        """
        super().__init__()
        self.address = address

    def send(self, msg):
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.settimeout(self.timeout)
            sock.connect(self.address)
            sock.send(msg)
            # sock.sendto(msg, self.address)


class QueuedSender(runnable.QueueHandler):
    """
    The UPD protocol is stateless but not necessarily thread-safe.

    ``QueuedSender`` uses a queue to send all UDP messages to one address
    from a new thread.

    """

    def __init__(self, address, **kwds):
        """
        :param str address: a pair (ip_address, port) to pass to socket.connect
        """
        super().__init__(**kwds)
        self.sender = Sender(address)

    def send(self, msg):
        self.sender.send(msg)


def sender(address, use_queue=True, **kwds):
    """
    :param str address: a pair (ip_address, port) to pass to socket.connect
    :param bool use_queue: if True, run the connection in a different thread
        with a queue
    """
    return QueuedSender(address, **kwds) if use_queue else Sender(address)


class Receiver(runnable.LoopThread):
    """
    Receive UDP messages in a thread
    """

    def __init__(self, address, bufsize=0x1000, receive=None, **kwds):
        super().__init__(**kwds)
        self.address = address
        self.bufsize = bufsize
        self.receive = receive or self.receive
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind(self.address)

    def __str__(self):
        return 'udp.Receiver(%s, %s)' % (self.address[0], hex(self.address[1]))

    def run_once(self):
        try:
            data, addr = self.socket.recvfrom(self.bufsize)
        except OSError as e:
            if e.errno != 9:
                raise
            if self.running:
                log.error('Someone else closed the socket')
                super().stop()
            return

        if data:
            self.receive(data)

    def stop(self):
        super().stop()
        try:
            self.socket.close()
        except Exception as e:
            log.error('Exception in socket.close: %s', e)


class QueuedReceiver(Receiver):
    """
    Receive UDP messages in a thread and put them on a queue.
    """

    def __init__(self, *args, **kwds):
        self.queue = queue.Queue()
        super().__init__(*args, **kwds)

    def receive(self, msg):
        self.queue.put(msg)



import os
from . import data_file


class PersistentDict(dict):
    """A dictionary that persists as a data_file on the filesystem.

    PersistentDict is constructed with a filename, which either does not exist,
    or contains YAML representing a previously stored value.

    Each time a PersistentDict is mutated, the file is rewritten with the new
    stored YAML data.
    """

    def __init__(self, filename):
        """
        :param c: the filename to store the DATA_FILE in
        """
        self.__filename = filename
        data = data_file.load(filename) if os.path.exists(filename) else {}
        super().__init__(data)

    def clear(self):
        super().clear()
        self.__write()

    def pop(self, *args):
        super().pop(*args)
        self.__write()

    def popitem(self):
        super().popitem()
        self.__write()

    def update(self, *args, **kwds):
        super().update(*args, **kwds)
        self.__write()

    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise ValueError('Keys for PersistentDict must be strings')
        super().__setitem__(key, value)
        self.__write()

    def __delitem__(self, key):
        super().__delitem__(key)
        self.__write()

    def __write(self):
        ordered = dict(sorted(self.items()))
        data_file.dump(ordered, self.__filename)



"""
Send an interrupt signal to a BiblioPixel process running on this
machine to kill it

DEPRECATED: use
.. code-block:: bash

    $ kill -int `bpa-pid`

"""

DESCRIPTION = """
Example:
`$ bp shutdown`

"""

from . restart import run, CHOICES


def set_parser(parser):
    parser.add_argument(
        'signal', nargs='?', default='SIGINT', choices=CHOICES,
        help='Signal to send.')

    parser.set_defaults(run=run)



import struct, webbrowser
from . import websocket
from .. server_driver import ServerDriver
from ... main import args

DEFAULT_SIMPIXEL_URL = 'http://simpixel.io'


class SimPixelOpenerServer(websocket.Server):
    def __init__(self, port, selectInterval):
        super().__init__(port, selectInterval)
        url = args.ARGS and (
            args.ARGS.simpixel or (args.ARGS.s and DEFAULT_SIMPIXEL_URL))
        if url is True:
            url = DEFAULT_SIMPIXEL_URL
        if url and not url.startswith('no'):
            webbrowser.open(url, new=0, autoraise=True)


class SimPixel(ServerDriver):
    """Output a simulation of your displat to the web-based
    simulator at http://simpixel.io

    Provides the same parameters of
    :py:class:`bibliopixel.drivers.driver_base.DriverBase` as
    well as those below:

    :param int port: Port to serve websocket server on.
    :param pixel_positions: Override the automatic generation of pyhsical
        pixel layout. This value can be generated via the methods in
        :py:mod:`bibliopixel.layout.geometry`
    """

    SERVER_CLASS = SimPixelOpenerServer
    SERVER_KWDS = {'selectInterval': 0.001}

    def __init__(self, num=1024, port=1337, **kwds):
        """
        Args:
            num:  number of LEDs being visualizer.
            port:  the port on which the SimPixel server is running.
            pixel_positions:  the positions of the LEDs in 3-d space.
            **kwds:  keywords passed to DriverBase.
        """
        super().__init__(num, address=port, **kwds)

    def _on_positions(self):
        # Flatten list of led positions.
        if self.server:
            pl = [c for p in self.pixel_positions for c in p]
            positions = bytearray(struct.pack('<%sh' % len(pl), *pl))
            self.server.update(positions=positions)

    def _send_packet(self):
        if not self.server:
            raise ValueError(
                'Tried to send a packet before Layout.start() was called')
        self.server.update(pixels=self._buf)


from ... util import deprecated
if deprecated.allowed():  # pragma: no cover
    DriverSimPixel = SimPixel



from .. util import log


class RETURN_CODES:
    SUCCESS = 255  # All is well
    REBOOT = 42  # Device reboot needed after configuration
    ERROR = 0  # Generic error
    ERROR_SIZE = 1  # Data received does not match given command length
    ERROR_UNSUPPORTED = 2  # Unsupported command
    ERROR_PIXEL_COUNT = 3  # Too many pixels for device
    ERROR_BAD_CMD = 4  # Unknown Command


RETURN_CODE_ERRORS = {
    RETURN_CODES.SUCCESS: 'Success!',
    RETURN_CODES.REBOOT: 'Device reboot needed after configuration.',
    RETURN_CODES.ERROR: 'Generic error',
    RETURN_CODES.ERROR_SIZE: 'Data packet size incorrect.',
    RETURN_CODES.ERROR_UNSUPPORTED: 'Unsupported configuration attempted.',
    RETURN_CODES.ERROR_PIXEL_COUNT: 'Wrong number of pixels for device.',
    RETURN_CODES.ERROR_BAD_CMD:
        'Unsupported protocol command. Check your device version.',
}


class BiblioSerialError(Exception):
    pass


def print_error(error):
    msg = RETURN_CODE_ERRORS.get(error, 'Unknown error occured.')
    log.error('%s: %s', error, msg)


def raise_error(error):
    raise BiblioSerialError(print_error(error))



"""
Update BiblioPixel's dependencies
"""

import os, sys, subprocess
from ... util import log


def run(args):
    root = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
    code = 0

    with open(os.path.join(root, 'requirements.txt')) as f:
        for requirement in f.read().splitlines():
            process = subprocess.Popen(
                ['pip', 'install', requirement, '--upgrade'],
                stdout=subprocess.PIPE)
            output, err = process.communicate()
            exit_code = process.wait()
            if exit_code:
                log.printer('ERROR upgrading', requirement, output, err)
                code = exit_code
            else:
                log.printer('Upgraded', requirement)

    sys.exit(code)


def set_parser(parser):
    parser.set_defaults(run=run)



import functools
from ...drivers.ledtype import LEDTYPE

USAGE = """
An LEDTYPE is represented by a string.

Possible LEDTYPEs are """ + ', '.join(LEDTYPE.__members__)


@functools.singledispatch
def make(c):
    raise ValueError("Don't understand type %s" % type(c), USAGE)


@make.register(LEDTYPE)
def _(c):
    return c


@make.register(str)
def _(c):
    return LEDTYPE[c]



class TriggerBase:
    def __init__(self, q, events):
        self.q = q
        self.events = events

    def trigger(self, name):
        self.q.put(
            {'req': 'trigger_animation', 'data': name, 'sender': 'Trigger'})



from . util import generate_header, even_dist, pointOnCircle, genVector

from . import deprecated
if deprecated.allowed():  # pragma: no cover
    from . attribute_dict import AttributeDict
    d = AttributeDict



"""
Run a bibliopixel demo
"""

DESCRIPTION = """
For the list of possible demos, type

.. code-block:: bash

    bp demo list

"""

import random

from . import args, common_flags, demo_table
from .. project import project
from .. util import log, pid_context

DEMO_OPTS = ', '.join(sorted(demo_table.DEMO_TABLE.keys()))


def make_runnable_animation(demo, args):
    if callable(demo):
        return demo(args).run

    if 'driver' in demo:
        if not demo['driver'].get('num'):
            if 'layout' in demo and demo['layout']['typename'] == 'cube':
                demo['driver']['num'] = args.width * args.height * args.depth
            else:
                demo['driver']['num'] = args.width * args.height

    if 'layout' in demo:
        layout = demo['layout']
        if 'width' in layout:
            layout['width'] = layout['width'] or args.width
        if 'x' in layout:
            layout['x'] = layout['x'] or args.width
        if 'height' in layout:
            layout['height'] = layout['height'] or args.height
        if 'y' in layout:
            layout['y'] = layout['y'] or args.height
        if 'z' in layout:
            layout['z'] = layout['z'] or args.depth

    project = common_flags.make_project(args, demo)
    return project.animation


def usage():
    log.printer('Available demos are: {}'.format(DEMO_OPTS))


def run(args):
    if args.name == 'list':
        usage()
        return

    args.simpixel = args.simpixel or True
    if not args.name:
        usage()
        args.name = random.choice(list(demo_table.DEMO_TABLE))
        log.printer('Randomly selected', args.name)

    try:
        demo = demo_table.DEMO_TABLE[args.name]
    except KeyError:
        raise KeyError('Unknown demo %s' % args.name)

    animation = make_runnable_animation(demo, args)
    animation.layout.start()
    animation.start()


def set_parser(parser):
    parser.set_defaults(run=run)
    common_flags.add_project_flags(parser)
    parser.add_argument(
        'name', nargs='?', default='',
        help='Name of demo to run. Options are: {}'.format(DEMO_OPTS))

    parser.add_argument(
        '--width', default=16, type=int,
        help='X dimension of display')

    parser.add_argument(
        '--height', default=16, type=int,
        help='Y dimension of display')

    parser.add_argument(
        '--depth', default=16, type=int,
        help='Z dimension of display. Only used for Cube demos.')



"""
Remove a project default file
"""
from ... project import defaults


def run(args):
    defaults.remove_defaults(args.name)


def set_parser(parser):
    parser.set_defaults(run=run)
    parser.add_argument('name', help='Remove a saved project defaults')



import socket, sys, time, os

from . driver_base import DriverBase
from .. import util
from .. util import log
from .. drivers.return_codes import RETURN_CODES


class CMDTYPE:
    SETUP_DATA = 1  # reserved for future use
    PIXEL_DATA = 2
    BRIGHTNESS = 3


class Network(DriverBase):
    """Driver for communicating with another device on the network.

    To be used with either
    BiblioPixelAnimations.receivers.GenericNetworkReceiver
    or :py:class:`.network_receiver.NetworkReceiver`

    Provides the same parameters of :py:class:`.driver_base.DriverBase` as
    well as those below:

    :param str host: Network hostname or IP address of receiver.
    :param int port: Network port of receiver
    """

    def __init__(self, num=0, width=0, height=0, host="localhost", port=3142,
                 **kwds):
        super().__init__(num, width, height, **kwds)

        self._host = host
        self._port = port

    def _connect(self):
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect((self._host, self._port))
            return s
        except socket.gaierror:
            error = "Unable to connect to or resolve host: {}".format(
                self._host)
            log.error(error)
            raise IOError(error)

    def _compute_packet(self):
        self._render()
        count = self.bufByteCount()
        self._packet = util.generate_header(CMDTYPE.PIXEL_DATA, count)
        self._packet.extend(self._buf)

    # Push new data to strand
    def _send_packet(self):
        try:
            s = self._connect()
            s.sendall(self._packet)

            resp = ord(s.recv(1))

            s.close()

            if resp != RETURN_CODES.SUCCESS:
                log.warning("Bytecount mismatch! %s", resp)

        except Exception as e:
            log.exception(e)
            error = "Problem communicating with network receiver!"
            log.error(error)
            raise IOError(error)

    def set_device_brightness(self, brightness):
        """Hardware specific method to set the global brightness for
        this driver's output. This method is required to be implemented,
        however, users should call
        :py:meth:`.driver_base.DriverBase.set_brightness`
        instead of calling this method directly.

        :param int brightness: 0-255 value representing the desired
            brightness level
        """
        packet = util.generate_header(CMDTYPE.BRIGHTNESS, 1)
        packet.append(self._brightness)
        s = self._connect()
        s.sendall(packet)
        resp = ord(s.recv(1))
        return resp == RETURN_CODES.SUCCESS


from .. util import deprecated
if deprecated.allowed():  # pragma: no cover
    DriverNetwork = Network



"""
Convert a string duration to a number of seconds.
"""
import re

SI_PREFIXES = (
    (10 ** 24, 'yotta', 'Y'),
    (10 ** 21, 'zetta', 'Z'),
    (10 ** 18, 'exa', 'E'),
    (10 ** 15, 'peta', 'P'),
    (10 ** 12, 'tera', 'T'),
    (10 ** 9, 'giga', 'G'),
    (10 ** 6, 'mega', 'M'),
    (10 ** 3, 'kilo', 'k'),
    (10 ** 2, 'hecto', 'h'),
    (10 ** 1, 'deka', 'da'),
    (10 ** -1, 'deci', 'd'),
    (10 ** -2, 'centi', 'c'),
    (10 ** -3, 'milli', 'm'),
    (10 ** -6, 'micro', 'µ'),
    (10 ** -9, 'nano', 'n'),
    (10 ** -12, 'pico', 'p'),
    (10 ** -15, 'femto', 'f'),
    (10 ** -18, 'atto', 'a'),
    (10 ** -21, 'zepto', 'z'),
    (10 ** -24, 'yocto', 'y'),
)

SI_SCALES = dict(
    [(p[1], p[0]) for p in SI_PREFIXES] +
    [(p[2], p[0]) for p in SI_PREFIXES] +
    [['u', 10 ** -9]])  # Cheat to allow usec

UNITS = {
    'week': 7 * 24 * 60 * 60,
    'day': 24 * 60 * 60,

    'hour': 60 * 60,
    'hr': 60 * 60,

    'minute': 60,
    'min': 60,

    'second': 1,
    'sec': 1,
    's': 1,
}

PART_MATCH = re.compile(r'([0-9.]+)([a-zA-Z]+)').fullmatch


def _get_units(s):
    # Some subtleties to handle the possibility of plural units.
    if s.endswith('ss'):
        raise ValueError('Unknown unit')

    def split_unit(s):
        for u in UNITS:
            if s.endswith(u):
                return s[:-len(u)], u

    prefix, unit = (
        s.endswith('s') and split_unit(s[:-1]) or
        split_unit(s) or
        (s, ''))

    if not unit:
        raise ValueError('Unknown unit ' + s)

    scale = UNITS[unit]

    if not prefix:
        return scale

    if prefix not in SI_SCALES:
        raise ValueError('Unknown metric prefix ' + prefix)

    return SI_SCALES[prefix] * scale


def parse(s):
    """
    Parse a string representing a time interval or duration into seconds,
    or raise an exception

    :param str s: a string representation of a time interval
    :raises ValueError: if ``s`` can't be interpreted as a duration

    """

    parts = s.replace(',', ' ').split()
    if not parts:
        raise ValueError('Cannot parse empty string')

    pieces = []
    for part in parts:
        m = PART_MATCH(part)
        pieces.extend(m.groups() if m else [part])

    if len(pieces) == 1:
        pieces.append('s')

    if len(pieces) % 2:
        raise ValueError('Malformed duration %s: %s: %s' % (s, parts, pieces))

    result = 0
    for number, units in zip(*[iter(pieces)] * 2):
        number = float(number)
        if number < 0:
            raise ValueError('Durations cannot have negative components')
        result += number * _get_units(units)

    return result



# Original code by msurguy:
# https://github.com/ManiacalLabs/BiblioPixel/issues/51#issuecomment-228662943

import os, sys
from . channel_order import ChannelOrder
from . driver_base import DriverBase
from .. util import log
from .. colors import gamma

WS_ERROR = """
PiWS281X Requires the rpi_ws281x C extension.

Install rpi_ws281x with the following shell commands:

    git clone https://github.com/jgarff/rpi_ws281x.git
    cd rpi_ws281x

    sudo apt-get install python-dev swig scons
    sudo scons

    cd python
    # If using default system python3
    sudo python3 setup.py build install
    # If using virtualenv, enter env then run
    python setup.py build install
"""

SUDO_ERROR = """
The PiWS281X driver needs to be run as sudo.  Rerun it with sudo, like this:
    sudo {command}
"""

try:
    from neopixel import Adafruit_NeoPixel, Color as NeoColor
except:
    NeoColor = None


PIN_CHANNEL = {
    12: 0,
    18: 0,
    40: 0,
    52: 0,
    13: 1,
    19: 1,
    41: 1,
    45: 1,
    53: 1,
    10: 0,  # Technically SPI
}

STRIP_TYPES = {
    3: 0x00100800,
    4: 0x18100800,
}


# Including follow as comment as URLs in docstrings don't play well with
# sphinx
# Discussion re: running as sudo
# https://groups.google.com/d/msg/maniacal-labs-users/6hV-2_-Xmqc/wmWJK709AQAJ
# https://github.com/jgarff/rpi_ws281x/blob/master/python/neopixel.py#L106


class PiWS281X(DriverBase):
    """
    Driver for controlling WS281X LEDs via the rpi_ws281x C-extension.
    Only supported on the Raspberry Pi 2, 3, and Zero

    This driver needs to be run as sudo and requires the rpi_ws281x C extension.

    Install rpi_ws281x with the following shell commands:

        git clone https://github.com/jgarff/rpi_ws281x.git
        cd rpi_ws281x

        sudo apt-get install python-dev swig scons
        sudo scons

        cd python
        # If using default system python3
        sudo python3 setup.py build install
        # If using virtualenv, enter env then run
        python setup.py build install

    Provides the same parameters of :py:class:`.driver_base.DriverBase` as
    well as those below:

    :param int gpio: GPIO pin to output to. Typically 18 or 13
    :param int ledFreqHz: WS2812B base data frequency in Hz. Only change to
        400000 if using very old WS218B LEDs
    :param int ledDma: DMA channel to use for generating signal
                       (between 1 and 14)
    :param bool ledInvert: True to invert the signal
                       (when using NPN transistor level shift)
    """

    def __init__(
            self, num, gamma=gamma.NEOPIXEL, c_order="RGB", gpio=18,
            ledFreqHz=800000, ledDma=5, ledInvert=False,
            color_channels=3, brightness=255, **kwds):

        if not NeoColor:
            raise ValueError(WS_ERROR)
        super().__init__(num, c_order=c_order, gamma=gamma, **kwds)
        self.gamma = gamma
        if gpio not in PIN_CHANNEL.keys():
            raise ValueError('{} is not a valid gpio option!')
        try:
            strip_type = STRIP_TYPES[color_channels]
        except:
            raise ValueError('In PiWS281X, color_channels can only be 3 or 4')

        self._strip = Adafruit_NeoPixel(
            num, gpio, ledFreqHz, ledDma, ledInvert, brightness,
            PIN_CHANNEL[gpio], strip_type)
        # Intialize the library (must be called once before other functions).
        try:
            self._strip.begin()
        except RuntimeError as e:
            if os.geteuid():
                if os.path.basename(sys.argv[0]) in ('bp', 'bibliopixel'):
                    command = ['bp'] + sys.argv[1:]
                else:
                    command = ['python'] + sys.argv
                error = SUDO_ERROR.format(command=' '.join(command))
                e.args = (error,) + e.args
            raise

    def set_brightness(self, brightness):
        self._strip.setBrightness(brightness)
        return True

    def _compute_packet(self):
        self._render()
        data = self._buf
        self._packet = [tuple(data[(p * 3):(p * 3) + 3])
                        for p in range(len(data) // 3)]

    def _send_packet(self):
        for i, p in enumerate(self._packet):
            self._strip.setPixelColor(i, NeoColor(*p))

        self._strip.show()



import os
import math
from . import errors
from ... util import log
from .. spi_interfaces import SPI_INTERFACES


class SpiBaseInterface(object):
    """ abstract class for different spi backends"""

    def __init__(self, dev, spi_speed):
        self._dev = dev
        self._spi_speed = spi_speed

    def send_packet(self, data):
        """SHOULD BE PRIVATE"""
        raise NotImplementedError

    def compute_packet(self, data):
        """SHOULD BE PRIVATE"""
        return data

    def error(self, text):
        """SHOULD BE PRIVATE"""
        msg = 'Error with dev: {}, spi_speed: {} - {}'.format(
            self._dev, self._spi_speed, text)
        log.error(msg)
        raise IOError(msg)


class SpiFileInterface(SpiBaseInterface):
    """ using os open/write to send data"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not os.path.exists(self._dev):
            self.error(errors.CANT_FIND_ERROR)

        self._spi = open(self._dev, 'wb')

        log.info('file io spi dev {:s}'.format(self._dev))

    def send_packet(self, data):
        """SHOULD BE PRIVATE"""
        package_size = 4032  # bit smaller than 4096 because of headers
        for i in range(int(math.ceil(len(data) / package_size))):
            start = i * package_size
            end = (i + 1) * package_size
            self._spi.write(data[start:end])
            self._spi.flush()


class SpiPeripheryInterface(SpiBaseInterface):
    """ using `python-periphery` to send data"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not os.path.exists(self._dev):
            self.error(errors.CANT_FIND_ERROR)

        try:
            from periphery import SPI
            self._spi = SPI(self._dev, 0, self._spi_speed * 1e6)
        except ImportError:
            self.error(errors.CANT_IMPORT_PERIPHERY_ERROR)

        log.info('periphery spi dev {:s} speed @ {:.2f} MHz'.format(
            self._dev, self._spi.max_speed / 1e6))

    def send_packet(self, data):
        """SHOULD BE PRIVATE"""
        package_size = 4032  # bit smaller than 4096 because of headers
        for i in range(int(math.ceil(len(data) / package_size))):
            start = i * package_size
            end = (i + 1) * package_size
            self._spi.transfer(data[start:end])


class SpiPyDevInterface(SpiBaseInterface):
    """ using py-spidev to send data"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        d = self._dev.replace('/dev/spidev', '')
        ids = (int(i) for i in d.split('.'))
        try:
            self._device_id, self._device_cs = ids
        except:
            self.error(errors.BAD_FORMAT_ERROR)

        if not os.path.exists(self._dev):
            self.error(errors.CANT_FIND_ERROR)
        # permissions check
        try:
            fd = open(self._dev, 'r')
            fd.close()
        except IOError as e:
            if e.errno == 13:
                self.error(errors.PERMISSION_ERROR)
            else:
                raise e
        # import spidev and cache error
        try:
            import spidev
            self._spi = spidev.SpiDev()
        except ImportError:
            self.error(errors.CANT_IMPORT_SPIDEV_ERROR)
        self._spi.open(self._device_id, self._device_cs)
        self._spi.max_speed_hz = int(self._spi_speed * 1e6)
        log.info(
            'py-spidev dev {:s} speed @ {:.2f} MHz'.format(
                self._dev, self._spi.max_speed_hz / 1e6))

    def send_packet(self, data):
        """SHOULD BE PRIVATE"""
        self._spi.xfer2(list(data))


class SpiDummyInterface(SpiBaseInterface):
    """ interface for testing proposal"""

    def send_packet(self, data):
        """ do nothing """
        pass


_SPI_INTERFACES = [
    SpiFileInterface,
    SpiPyDevInterface,
    SpiPeripheryInterface,
    SpiDummyInterface
]



from . animation import BaseAnimation
from .. util import colors


class Fill(BaseAnimation):
    """
    Fill the screen with a single color.
    """
    def __init__(self, *args, color='black', **kwds):
        super().__init__(*args, preclear=False, **kwds)

        is_numpy = hasattr(self.color_list, 'dtype')
        self._set_color = self._set_numpy if is_numpy else self._set_classic
        self._color = color

    def pre_run(self):
        self.color = self._color

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, color):
        self._color = colors.to_color(color)
        self._set_color()

    def _set_numpy(self):
        self.color_list[:None] = self._color

    def _set_classic(self):
        self.color_list[:] = [self._color] * len(self.color_list)



from .. util import colors
from . layout import MultiLayout
from . geometry import make_strip_coord_map_multi
from . geometry.strip import make_strip_coord_map_positions


class Strip(MultiLayout):
    CLONE_ATTRS = MultiLayout.CLONE_ATTRS + ('brightness', 'pixelWidth')

    def __init__(self, drivers, threadedUpdate=False,
                 brightness=255, pixelWidth=1, **kwargs):
        self.gen_multi = make_strip_coord_map_multi
        super().__init__(drivers, threadedUpdate, brightness, **kwargs)

        self.pixelWidth = pixelWidth
        if self.pixelWidth < 1 or self.pixelWidth > self.numLEDs:
            raise ValueError(
                "pixelWidth must be greater than 0 and less than or equal to "
                "the total LED count!")
        if self.numLEDs % self.pixelWidth != 0:
            raise ValueError(
                "Total LED count must be evenly divisible by pixelWidth!")
        if self.pixelWidth == 1:
            self.set = self._set
        else:
            self.set = self._setScaled
            self.num_pixels = int(self.numLEDs / self.pixelWidth)

        if self.coord_map:
            if len(self.coord_map) != self.numLEDs:
                raise ValueError(
                    'coord_map length must equal total number of pixels!')
            self.set_base = self._set_strip_mapped
        else:
            self.set_base = self._set_base

        self.set_pixel_positions(make_strip_coord_map_positions(self.numLEDs))

    @property
    def shape(self):
        """Returns a 1-tuple with the length of the strip"""
        return self.numLEDs,

    def get(self, pixel):
        """Get RGB color tuple of color at index pixel"""
        return self._get_base(pixel * self.pixelWidth)

    def set(self, pixel, color):
        """
        Set the pixel color at position x in the strip.
        """
        # The actual implementation of this method is computed at construction
        # time and monkey-patched in from one of self._setTexture,
        # self.__setNormal or self.__setScaled
        raise NotImplementedError

    def _set_strip_mapped(self, pixel, color):
        if pixel >= 0 and pixel < self.numLEDs:
            pixel = self.coord_map[pixel]
            self._set_base(pixel, color)

    # Set single pixel to Color value
    def _set(self, pixel, color):
        """Set pixel to RGB color tuple"""
        self.set_base(pixel, color)

    def _setScaled(self, pixel, color):
        start = pixel * self.pixelWidth
        for p in range(start, start + self.pixelWidth):
            self.set_base(p, color)

    # Set single pixel to RGB value
    def setRGB(self, pixel, r, g, b):
        """Set single pixel using individual RGB values instead of tuple"""
        self.set(pixel, (r, g, b))

    def setHSV(self, pixel, hsv):
        """Set single pixel to HSV tuple"""
        color = colors.hsv2rgb(hsv)
        self.set(pixel, color)

    # turns off the desired pixel
    def setOff(self, pixel):
        """Set single pixel off"""
        self.set(pixel, (0, 0, 0))


from .. util import deprecated
if deprecated.allowed():  # pragma: no cover
    LEDStrip = Strip



import bibliopixel, flask, logging, os, queue, sys
from ... util import log

MESSAGE = """\
Remote UI Server available at: {1}
Local: http://localhost:{0}"""

EXTERNAL_ACCESS_MESSAGE = MESSAGE + """
External: http://<system_ip>:{0}"""

if '--verbose' not in sys.argv:
    logging.getLogger('werkzeug').setLevel(logging.ERROR)

ROOT_DIR = os.path.dirname(os.path.dirname(bibliopixel.__file__))
STATIC_DIR = os.path.join(ROOT_DIR, 'ui', 'web_remote')


class RemoteServer:
    def __init__(self, q_send, q_recv):
        assert os.path.exists(STATIC_DIR)
        self.app = flask.Flask('BP Remote', static_folder=STATIC_DIR)
        self._set_routes()
        self.q_send = q_send
        self.q_recv = q_recv

    @classmethod
    def shutdown_server(cls):
        log.warning('Shutting down UI server.')
        func = flask.request.environ.get('werkzeug.server.shutdown')
        if func is None:
            return False
        func()
        return True

    def _set_routes(self):
        self.app.route('/')(self.index)
        self.app.route('/<path:path>')(self.static_files)
        self.app.route('/run_animation/<string:animation>')(self.run_animation)
        self.app.route('/stop')(self.stop_animation)
        self.app.route('/api/<string:request>')(self.api)
        self.app.route('/api/<string:request>/<data>')(self.api)

    def index(self):
        log.debug('index.html')
        return self.app.send_static_file('index.html')

    def static_files(self, path):
        log.debug('static_files: %s', path)
        return self.app.send_static_file(path)

    def run_animation(self, animation):
        return self.api('run_animation', data=animation)

    def stop_animation(self):
        return self.api('stop_animation')

    def api(self, request, data=None):
        log.debug('api: %s, %s', request, data)
        msg = {
            'req': request.lower(),
            'data': data,
            'sender': 'RemoteServer'}

        self.q_send.put(msg)

        try:
            status, data = self.q_recv.get(timeout=5)
        except queue.Empty:
            status, data = False, 'Timeout waiting for response.'

        return flask.jsonify({
            'status': status,
            'msg': 'OK' if status else data,
            'data': data if status else None,
        })


def run_server(external_access, port, q_send, q_recv):
    server = RemoteServer(q_recv, q_send)
    if external_access:
        host_ip = '0.0.0.0'
        msg = EXTERNAL_ACCESS_MESSAGE
    else:
        host_ip = 'localhost'
        msg = MESSAGE

    log.info(msg.format(port, host_ip))

    import werkzeug.serving
    werkzeug.serving.run_simple(host_ip, port, server.app, threaded=True)



import traceback
from .. util import log
from . import animation


class Failed(animation.Animation):
    """
    An animation that's created when we fail to load or construct the
    animation that was originally specified
    """

    def __init__(self, layout, desc, exception):
        super().__init__(layout)
        self._set_runner({})
        log.error('Unable to create animation for %s', desc)
        debug = log.get_log_level() <= log.DEBUG
        if debug:
            try:
                msg = traceback.format_exc()
            except:
                msg = str(exception)
        else:
            msg = str(exception)

        log.error('\n%s', msg)
        self.desc = desc
        self.exception = exception
        self.empty = True



def check(kwds, name):
    if kwds:
        msg = ', '.join('"%s"' % s for s in sorted(kwds))
        s = '' if len(kwds) == 1 else 's'
        raise ValueError('Unknown attribute%s for %s: %s' % (s, name, msg))


def set_reserved(value, section, name=None, data=None, **kwds):
    check(kwds, '%s %s' % (section, value.__class__.__name__))
    value.name = name
    value.data = data



import copy
from . import parallel
from .. util import color_list


class Mixer(parallel.Parallel):
    def __init__(self, *args, levels=None, master=1, **kwds):
        self.master = master

        super().__init__(*args, **kwds)
        self.mixer = color_list.Mixer(
            self.color_list,
            [a.color_list for a in self.animations],
            levels)

    @property
    def levels(self):
        return self.mixer.levels

    @levels.setter
    def levels(self, levels):
        self.mixer.levels[:] = levels

    def step(self, amt=1):
        super().step(amt)
        self.mixer.clear()
        self.mixer.mix(self.master)



"""
Fill in missing parts of a project and cleans up its sections
so they have exactly the right constructor.
"""

import copy
from . import aliases, alias_lists, construct, merge
from .. import layout, util
from .. animation.strip import BaseStripAnim

DEFAULT_DRIVERS = [construct.to_type('simpixel')]


def fill(desc):
    desc = _fill_aliases(desc)
    desc = _fill_animation(desc)
    desc = _fill_drivers(desc)
    desc = _fill_shape(desc)  # Must come after fill_drivers
    desc = _fill_numbers(desc)
    desc = _fill_controls(desc)

    return desc


def fill_layout(animation):
    # Try to fill in the layout if it's missing.
    # This is called from Project.__init__ - TODO: fix that.
    datatype = animation['datatype']

    args = getattr(datatype, 'LAYOUT_ARGS', BaseStripAnim.LAYOUT_ARGS)
    layout_cl = getattr(datatype, 'LAYOUT_CLASS', BaseStripAnim.LAYOUT_CLASS)

    args = {k: animation[k] for k in args if k in animation}
    return dict(args, datatype=layout_cl)


def _fill_aliases(desc):
    def unalias(k):
        if any(k.startswith(m) for m in aliases.ALIAS_MARKERS):
            return k[1:]
        return k

    al = desc.pop('aliases', {})
    alias_lists.PROJECT_ALIASES = {unalias(k): v for k, v in al.items()}
    return desc


def _fill_animation(desc):
    if not desc['animation']:
        raise ValueError('Missing "animation" section')

    desc['animation'] = construct.to_type_constructor(
        desc['animation'], 'bibliopixel.animation')
    datatype = desc['animation'].get('datatype')
    if not datatype:
        e = desc['animation'].get('_exception')
        raise e or ValueError('Missing "datatype" in "animation" section')

    desc['animation'].setdefault('name', datatype.__name__)

    from .. animation import sequence
    if not issubclass(datatype, sequence.Sequence):
        # Magic here to allow this to work for non-sequences.
        length = desc['animation'].pop('length', [])
        if length:
            desc['run']['seconds'] = length[0]

    desc = merge.merge(getattr(datatype, 'PROJECT', {}), desc)

    run = desc.pop('run')
    anim_run = desc['animation'].setdefault('run', {})
    if run:
        desc['animation']['run'] = dict(run, **anim_run)
    return desc


def _fill_drivers(desc):
    driver = construct.to_type(desc.pop('driver', {}))
    drivers = [construct.to_type(d) for d in desc['drivers']]
    if driver:
        if drivers:
            drivers = [dict(driver, **d) for d in drivers]
        else:
            drivers = [driver]

    desc['drivers'] = drivers or DEFAULT_DRIVERS
    return desc


def _fill_shape(desc):
    desc = copy.deepcopy(desc)
    dimensions = desc.pop('dimensions', None)
    if dimensions:
        util.deprecated.deprecated('Project section "dimensions"')

    shape = desc.pop('shape', None) or dimensions
    if not shape:
        return desc

    if len(desc['drivers']) != 1:
        raise ValueError('Cannot use dimensions with more than one driver')

    if isinstance(shape, int):
        shape = [shape]
    elif not isinstance(shape, list):
        raise ValueError('`shape` must be a number or a list, was "%s"' % shape)

    ldesc = construct.to_type_constructor(desc.get('layout') or {},
                                          python_path='bibliopixel.layout')
    driver = desc['drivers'][0]

    if len(shape) == 1:
        driver['num'] = shape[0]
        ldesc.setdefault('datatype', layout.strip.Strip)

    elif len(shape) == 2:
        driver['num'] = shape[0] * shape[1]
        ldesc.setdefault('datatype', layout.matrix.Matrix)
        ldesc.update(width=shape[0], height=shape[1])

    elif len(shape) == 3:
        driver['num'] = shape[0] * shape[1] * shape[2]
        ldesc.setdefault('datatype', layout.cube.Cube)
        ldesc.update(x=shape[0], y=shape[1], z=shape[2])

    else:
        raise ValueError('Dimension %s > 3' % len(shape))

    desc['layout'] = ldesc
    return desc


def _fill_numbers(desc):
    numbers = desc.pop('numbers', 'python')
    if numbers != 'python':
        desc.setdefault('maker', {})['numpy_dtype'] = numbers

    return desc


def _fill_controls(desc):
    controls = desc.get('controls', None)
    if isinstance(controls, (str, dict)):
        desc['controls'] = [controls]
    return desc



import contextlib


@contextlib.contextmanager
def add(*args):
    """
    A context manager that appends arguments to any exception thrown

    :param args: Arguments to be appended to the ``.args`` attribute of any
                 exception that is thrown while the context manager is active
    """
    try:
        yield
    except Exception as e:
        e.args = args + e.args
        raise



class Gamma(object):
    """
    Compute a fixed gamma table with 256 entries.
    """

    def __init__(self, gamma=1.0, offset=0, lower_bound=0):
        """
        :param float gamma: the root for gamma computation
        :param float offset: an offset added to the result
        :param int lower_bound: The lowest possible output value - the highest
                                is always 255.

        """
        self.gamma = gamma
        self.offset = offset
        self.lower_bound = lower_bound
        width = 255 - lower_bound

        def gam(i):
            return int(lower_bound + pow(i / 255, gamma) * width + offset)

        self.table = tuple(gam(i) for i in range(256))

    def get(self, i):
        """
        :returns: The gamma table entry
        :param int i: the index into the table
        """
        return self.table[max(0, min(255, int(i)))]


# From https://github.com/scottjgibson/PixelPi/blob/master/pixelpi.py
APA102 = Gamma(gamma=2.5, offset=0.5)
WS2801 = SM16716 = Gamma(gamma=2.5)
NONE = DEFAULT = Gamma()  # For when you need no correction

# From http://rgb-123.com/ws2812-color-output/
WS2812B = NEOPIXEL = WS2812 = Gamma(gamma=1 / 0.45)

# Color calculations from
# http://learn.adafruit.com/light-painting-with-raspberry-pi
LPD8806 = Gamma(gamma=2.5, offset=0.5, lower_bound=128)



import functools, numbers
from ... util.colors import gamma

NAMES = [d for d in dir(gamma) if d.isupper()]

USAGE = """
A Gamma table is represented by one of:

* A number, representing a floating gamma correction value
* A string, representing a name of a standard gamma function
* A list of up to three elements: arguments to Gamma's constructor.
* A dictionary representing arguments to Gamma's constructor.

Possible string names  are """ + ', '.join(NAMES)


@functools.singledispatch
def make(c):
    raise ValueError("Don't understand type %s" % type(c), USAGE)


@make.register(numbers.Number)
def _(c):
    return gamma.Gamma(c)


@make.register(str)
def _(c):
    if c not in NAMES:
        raise ValueError('Unknown gamma %s: valid names are %s' % (c, NAMES))

    return getattr(gamma, c)


@make.register(list)
@make.register(tuple)
def _(c):
    return gamma.Gamma(*c)


@make.register(dict)
def _(c):
    return gamma.Gamma(**c)



import functools, queue, traceback
from .. util import log


class EditQueue(queue.Queue):
    """
    ``Edits``s are thread-unsafe operations to be executed later at a
    thread-safe time.  In the animation loop, this happens in
    :py:method:`bibliopixel.animation.Animation.preframe_callback`.

    The typical `Edit` is a change to some parameter of an animation,
    but in general is represented by a call to a function with fixed parameters.
    """

    def put_edit(self, f, *args, **kwds):
        """
        Defer an edit to run on the EditQueue.

        :param callable f: The function to be called
        :param tuple args: Positional arguments to the function
        :param tuple kwds: Keyword arguments to the function
        :throws queue.Full: if the queue is full
        """
        self.put_nowait(functools.partial(f, *args, **kwds))

    def get_and_run_edits(self):
        """
        Get all the edits in the queue, then execute them.

        The algorithm gets all edits, and then executes all of them.  It does
        *not* pull off one edit, execute, repeat until the queue is empty, and
        that means that the queue might not be empty at the end of
        ``run_edits``, because new edits might have entered the queue
        while the previous edits are being executed.

        This has the advantage that if edits enter the queue faster than they
        can be processed, ``get_and_run_edits`` won't go into an infinite loop,
        but rather the queue will grow unboundedly, which that can be
        detected, and mitigated and reported on - or if Queue.maxsize is
        set, `bp` will report a fairly clear error and just dump the edits
        on the ground.
        """
        if self.empty():
            return

        edits = []
        while True:
            try:
                edits.append(self.get_nowait())
            except queue.Empty:
                break

        for e in edits:
            try:
                e()
            except:
                log.error('Error on edit %s', e)
                traceback.print_exc()



import re
from .. util import log
from . import alias_lists

ALIAS_MARKERS = '@$'
SEPARATORS = re.compile(r'([./#]|[^./#]+)')


def resolve(typename, aliases=None):
    aliases = aliases or {}

    def get(s):
        return aliases.get(s) or alias_lists.get_alias(s) or s

    def get_all(typename):
        for part in SEPARATORS.split(typename):
            is_alias = part and (part[0] in ALIAS_MARKERS)
            yield get(part[1:]) if is_alias else part

    return ''.join(get_all(get(typename)))



import collections, json, os, sys, yaml

# Allow open to be patched for tests.
open = __builtins__['open']

ALWAYS_LOAD_YAML, ALWAYS_DUMP_YAML = True, True


def dumps(data, use_yaml=None, safe=True, **kwds):
    """
    Dumps data into a nicely formatted JSON string.

    :param dict data: a dictionary to dump
    :param kwds: keywords to pass to json.dumps
    :returns: a string with formatted data
    :rtype: str
    """
    if use_yaml is None:
        use_yaml = ALWAYS_DUMP_YAML

    if use_yaml:
        dumps = yaml.safe_dump if safe else yaml.dump
    else:
        dumps = json.dumps
        kwds.update(indent=4, sort_keys=True)
        if not safe:
            kwds.update(default=repr)
    return dumps(data, **kwds)


def dump(data, file=sys.stdout, use_yaml=None, **kwds):
    """
    Dumps data as nicely formatted JSON string to a file or file handle

    :param dict data: a dictionary to dump
    :param file: a filename or file handle to write to
    :param kwds: keywords to pass to json.dump
    """
    if use_yaml is None:
        use_yaml = ALWAYS_DUMP_YAML

    def dump(fp):
        if use_yaml:
            yaml.safe_dump(data, stream=fp, **kwds)
        else:
            json.dump(data, fp, indent=4, sort_keys=True, **kwds)

    if not isinstance(file, str):
        return dump(file)

    if os.path.isabs(file):
        parent = os.path.dirname(file)
        if not os.path.exists(parent):
            os.makedirs(parent, exist_ok=True)

    with open(file, 'w') as fp:
        return dump(fp)


def loads(s, use_yaml=None, filename=''):
    if use_yaml is None:
        use_yaml = ALWAYS_LOAD_YAML

    if not (filename.endswith('.yml') or use_yaml):
        return json.loads(s)

    def fix(d):
        if isinstance(d, dict):
            return {str(k): fix(v) for k, v in d.items()}
        if isinstance(d, list):
            return [fix(i) for i in d]
        if not isinstance(d, (int, float, bool, str, type(None))):
            raise ValueError('Wrong type %s' % type(d))
        return d

    return fix(yaml.safe_load(s))


def load(file, use_yaml=None):
    """
    Loads not only JSON files but also YAML files ending in .yml.

    :param file: a filename or file handle to read from
    :returns: the data loaded from the JSON or YAML file
    :rtype: dict
    """
    if isinstance(file, str):
        fp = open(file)
        filename = file
    else:
        fp = file
        filename = getattr(fp, 'name', '')

    try:
        return loads(fp.read(), use_yaml, filename)

    except Exception as e:
        e.args = ('There was a error in the data file', filename) + e.args
        raise



"""
Handle DMX and MIDI channel offsets
"""


class OffsetRange:
    def __init__(self, offset=0, begin=None, end=None):
        """
        Unlike a `range`, an OffsetRange includes both its begin *and* its end,
        so it's closer to how regular people think of a range - for example
        that DMX channels are in the range 1-512.
        """

        self.begin = self.BEGIN if begin is None else begin
        self.end = self.END if end is None else end

        if not (self.BEGIN <= self.begin <= self.end <= self.END):
            raise ValueError('Bad range: FAILED %s <= %s <=  %s <= %s' %
                             (self.BEGIN, self.begin, self.end, self.END))

        self.offset = offset

    def full_range(self):
        return self.END - self.BEGIN + 1

    def index(self, i, length=None):
        """Return an integer index or None"""
        if self.begin <= i <= self.end:
            index = i - self.BEGIN - self.offset
            if length is None:
                length = self.full_range()
            else:
                length = min(length, self.full_range())

            if 0 <= index < length:
                return index

    def read_from(self, data, pad=0):
        """
        Returns a generator with the elements "data" taken by offset, restricted
        by self.begin and self.end, and padded on either end by `pad` to get
        back to the original length of `data`
        """
        for i in range(self.BEGIN, self.END + 1):
            index = self.index(i, len(data))
            yield pad if index is None else data[index]

    def copy_to(self, source, target):
        if not target:
            return

        offset = self.offset
        if offset < 0:
            source = source[-offset:]
            offset = 0

        begin = min(self.begin - self.BEGIN, len(target) - 1)
        end = min(self.end - self.BEGIN, len(target) - 1)
        length = min(len(source) - offset, end - begin + 1)

        target[begin:begin + length] = source[offset:offset + length]

    @classmethod
    def make(cls, x=0):
        try:
            return cls(**x)
        except TypeError:
            return cls(offset=x)


class MidiChannel(OffsetRange):
    BEGIN = 1
    END = 16


class MidiNote(OffsetRange):
    BEGIN = 0
    END = 127


class DMXChannel(OffsetRange):
    BEGIN = 1
    END = 512



import sys
from . import args, commands, common_flags
from .. util import log
from .. project import aliases, project

__all__ = ['main']


def main():
    args.parse_args(commands, common_flags)
    run = getattr(args.ARGS, 'run', None)
    if not run:
        log.printer('ERROR: No command entered')
        log.printer('Valid commands are:')
        log.printer('    ', ', '.join(commands.COMMANDS))
        log.printer()
        log.printer('For more help, type')
        log.printer()
        log.printer('    bp --help')
        sys.exit(-1)

    common_flags.execute_args(args.ARGS)

    try:
        result = run(args.ARGS) or 0
    except Exception as e:
        if args.ARGS.verbose:
            raise
        log.printer('ERROR:', e.args and e.args[0], file=sys.stderr)
        log.printer(*e.args[1:], sep='\n', file=sys.stderr)
        result = getattr(e, 'errorcode', -1)

    sys.exit(result)



import math
from .. util import colors, log
from . import font

##########################################################################
# Drawing Functions
# Lovingly borrowed from Adafruit
# https://github.com/adafruit/Adafruit-GFX-Library/blob/master/Adafruit_GFX.cpp
##########################################################################


def draw_circle(setter, x0, y0, r, color=None):
    """
    Draws a circle at point x0, y0 with radius r of the specified RGB color
    """
    f = 1 - r
    ddF_x = 1
    ddF_y = -2 * r
    x = 0
    y = r

    setter(x0, y0 + r, color)
    setter(x0, y0 - r, color)
    setter(x0 + r, y0, color)
    setter(x0 - r, y0, color)

    while x < y:
        if f >= 0:
            y -= 1
            ddF_y += 2
            f += ddF_y
        x += 1
        ddF_x += 2
        f += ddF_x

        setter(x0 + x, y0 + y, color)
        setter(x0 - x, y0 + y, color)
        setter(x0 + x, y0 - y, color)
        setter(x0 - x, y0 - y, color)
        setter(x0 + y, y0 + x, color)
        setter(x0 - y, y0 + x, color)
        setter(x0 + y, y0 - x, color)
        setter(x0 - y, y0 - x, color)


def _draw_circle_helper(setter, x0, y0, r, cornername, color=None):
    f = 1 - r
    ddF_x = 1
    ddF_y = -2 * r
    x = 0
    y = r

    while x < y:
        if (f >= 0):
            y -= 1
            ddF_y += 2
            f += ddF_y
        x += 1
        ddF_x += 2
        f += ddF_x
        if (cornername & 0x4):
            setter(x0 + x, y0 + y, color)
            setter(x0 + y, y0 + x, color)

        if (cornername & 0x2):
            setter(x0 + x, y0 - y, color)
            setter(x0 + y, y0 - x, color)

        if (cornername & 0x8):
            setter(x0 - y, y0 + x, color)
            setter(x0 - x, y0 + y, color)

        if (cornername & 0x1):
            setter(x0 - y, y0 - x, color)
            setter(x0 - x, y0 - y, color)


def _fill_circle_helper(setter, x0, y0, r, cornername, delta, color=None):
    f = 1 - r
    ddF_x = 1
    ddF_y = -2 * r
    x = 0
    y = r

    while (x < y):
        if (f >= 0):
            y -= 1
            ddF_y += 2
            f += ddF_y
        x += 1
        ddF_x += 2
        f += ddF_x

        if (cornername & 0x1):
            _draw_fast_vline(setter, x0 + x, y0 - y, 2 * y + 1 + delta, color)
            _draw_fast_vline(setter, x0 + y, y0 - x, 2 * x + 1 + delta, color)

        if (cornername & 0x2):
            _draw_fast_vline(setter, x0 - x, y0 - y, 2 * y + 1 + delta, color)
            _draw_fast_vline(setter, x0 - y, y0 - x, 2 * x + 1 + delta, color)


def fill_circle(setter, x0, y0, r, color=None):
    """Draws a filled circle at point x0,y0 with radius r and specified color"""
    _draw_fast_vline(setter, x0, y0 - r, 2 * r + 1, color)
    _fill_circle_helper(setter, x0, y0, r, 3, 0, color)


def draw_line(setter, x0, y0, x1, y1, color=None, colorFunc=None, aa=False):
    if aa:
        wu_line(setter, x0, y0, x1, y1, color, colorFunc)
    else:
        bresenham_line(setter, x0, y0, x1, y1, color, colorFunc)


# Bresenham's algorithm
def bresenham_line(setter, x0, y0, x1, y1, color=None, colorFunc=None):
    """Draw line from point x0,y0 to x,1,y1. Will draw beyond matrix bounds."""
    steep = abs(y1 - y0) > abs(x1 - x0)
    if steep:
        x0, y0 = y0, x0
        x1, y1 = y1, x1

    if x0 > x1:
        x0, x1 = x1, x0
        y0, y1 = y1, y0

    dx = x1 - x0
    dy = abs(y1 - y0)

    err = dx / 2

    if y0 < y1:
        ystep = 1
    else:
        ystep = -1

    count = 0
    for x in range(x0, x1 + 1):
        if colorFunc:
            color = colorFunc(count)
            count += 1

        if steep:
            setter(y0, x, color)
        else:
            setter(x, y0, color)

        err -= dy
        if err < 0:
            y0 += ystep
            err += dx
# END Bresenham's algorithm


# Xiaolin Wu's _line Algorithm
def wu_line(setter, x0, y0, x1, y1, color=None, colorFunc=None):
    funcCount = [0]  # python2 hack since nonlocal not available

    def plot(x, y, level):
        c = color
        if colorFunc:
            c = colorFunc(funcCount[0])
            funcCount[0] += 1

        c = colors.color_scale(color, int(255 * level))
        setter(int(x), int(y), c)

    def ipart(x):
        return int(x)

    def fpart(x):
        return x - math.floor(x)

    def rfpart(x):
        return 1.0 - fpart(x)

    steep = abs(y1 - y0) > abs(x1 - x0)
    if steep:
        x0, y0 = y0, x0
        x1, y1 = y1, x1

    if x0 > x1:
        x0, x1 = x1, x0
        y0, y1 = y1, y0

    dx = x1 - x0
    dy = y1 - y0
    gradient = dy / dx

    # handle first endpoint
    xend = round(x0)
    yend = y0 + gradient * (xend - x0)
    xgap = rfpart(x0 + 0.5)
    xpxl1 = xend  # this will be used in the main loop
    ypxl1 = ipart(yend)

    if steep:
        plot(ypxl1, xpxl1, rfpart(yend) * xgap)
        plot(ypxl1 + 1, xpxl1, fpart(yend) * xgap)
    else:
        plot(xpxl1, ypxl1, rfpart(yend) * xgap)
        plot(xpxl1, ypxl1 + 1, fpart(yend) * xgap)

    # first y-intersection for the main loop
    intery = yend + gradient

    # handle second endpoint
    xend = round(x1)
    yend = y1 + gradient * (xend - x1)
    xgap = fpart(x1 + 0.5)
    xpxl2 = xend  # this will be used in the main loop
    ypxl2 = ipart(yend)

    if steep:
        plot(ypxl2, xpxl2, rfpart(yend) * xgap)
        plot(ypxl2 + 1, xpxl2, fpart(yend) * xgap)
    else:
        plot(xpxl2, ypxl2, rfpart(yend) * xgap)
        plot(xpxl2, ypxl2 + 1, fpart(yend) * xgap)

    # main loop
    for x in range(int(xpxl1 + 1), int(xpxl2)):
        if steep:
            plot(ipart(intery), x, rfpart(intery))
            plot(ipart(intery) + 1, x, fpart(intery))
        else:
            plot(x, ipart(intery), rfpart(intery))
            plot(x, ipart(intery) + 1, fpart(intery))
        intery = intery + gradient

# END Xiaolin Wu's _line Algorithm


def _draw_fast_vline(setter, x, y, h, color=None, aa=False):
    draw_line(setter, x, y, x, y + h - 1, color, aa)


def _draw_fast_hline(setter, x, y, w, color=None, aa=False):
    draw_line(setter, x, y, x + w - 1, y, color, aa)


def draw_rect(setter, x, y, w, h, color=None, aa=False):
    """Draw rectangle with top-left corner at x,y, width w and height h"""
    _draw_fast_hline(setter, x, y, w, color, aa)
    _draw_fast_hline(setter, x, y + h - 1, w, color, aa)
    _draw_fast_vline(setter, x, y, h, color, aa)
    _draw_fast_vline(setter, x + w - 1, y, h, color, aa)


def fill_rect(setter, x, y, w, h, color=None, aa=False):
    """Draw solid rectangle with top-left corner at x,y, width w and height h"""
    for i in range(x, x + w):
        _draw_fast_vline(setter, i, y, h, color, aa)


def draw_round_rect(setter, x, y, w, h, r, color=None, aa=False):
    """Draw rectangle with top-left corner at x,y, width w, height h,
    and corner radius r.
    """
    _draw_fast_hline(setter, x + r, y, w - 2 * r, color, aa)  # Top
    _draw_fast_hline(setter, x + r, y + h - 1, w - 2 * r, color, aa)  # Bottom
    _draw_fast_vline(setter, x, y + r, h - 2 * r, color, aa)  # Left
    _draw_fast_vline(setter, x + w - 1, y + r, h - 2 * r, color, aa)  # Right
    # draw four corners
    _draw_circle_helper(setter, x + r, y + r, r, 1, color, aa)
    _draw_circle_helper(setter, x + w - r - 1, y + r, r, 2, color, aa)
    _draw_circle_helper(setter, x + w - r - 1, y + h - r - 1, r, 4, color, aa)
    _draw_circle_helper(setter, x + r, y + h - r - 1, r, 8, color, aa)


def fill_round_rect(setter, x, y, w, h, r, color=None, aa=False):
    """Draw solid rectangle with top-left corner at x,y, width w, height h,
    and corner radius r"""
    fill_rect(setter, x + r, y, w - 2 * r, h, color, aa)
    _fill_circle_helper(setter, x + w - r - 1, y + r, r,
                        1, h - 2 * r - 1, color, aa)
    _fill_circle_helper(setter, x + r, y + r, r, 2, h - 2 * r - 1, color, aa)


def draw_triangle(setter, x0, y0, x1, y1, x2, y2, color=None, aa=False):
    """Draw triangle with points x0,y0 - x1,y1 - x2,y2"""
    draw_line(setter, x0, y0, x1, y1, color, aa)
    draw_line(setter, x1, y1, x2, y2, color, aa)
    draw_line(setter, x2, y2, x0, y0, color, aa)


def fill_triangle(setter, x0, y0, x1, y1, x2, y2, color=None, aa=False):
    """Draw solid triangle with points x0,y0 - x1,y1 - x2,y2"""
    a = b = y = last = 0

    if y0 > y1:
        y0, y1 = y1, y0
        x0, x1 = x1, x0
    if y1 > y2:
        y2, y1 = y1, y2
        x2, x1 = x1, x2
    if y0 > y1:
        y0, y1 = y1, y0
        x0, x1 = x1, x0

    if y0 == y2:  # Handle awkward all-on-same-line case as its own thing
        a = b = x0
        if x1 < a:
            a = x1
        elif x1 > b:
            b = x1
        if x2 < a:
            a = x2
        elif x2 > b:
            b = x2
            _draw_fast_hline(setter, a, y0, b - a + 1, color, aa)

    dx01 = x1 - x0
    dy01 = y1 - y0
    dx02 = x2 - x0
    dy02 = y2 - y0
    dx12 = x2 - x1
    dy12 = y2 - y1
    sa = 0
    sb = 0

    # For upper part of triangle, find scanline crossings for segments
    # 0-1 and 0-2.  If y1=y2 (flat-bottomed triangle), the scanline y1
    # is included here (and second loop will be skipped, avoiding a /0
    # error there), otherwise scanline y1 is skipped here and handled
    # in the second loop...which also avoids a /0 error here if y0=y1
    # (flat-topped triangle).

    if y1 == y2:
        last = y1  # include y1 scanline
    else:
        last = y1 - 1  # skip it

    for y in range(y, last + 1):
        a = x0 + sa / dy01
        b = x0 + sb / dy02
        sa += dx01
        sb += dx02

        if a > b:
            a, b = b, a
            _draw_fast_hline(setter, a, y, b - a + 1, color, aa)

    # For lower part of triangle, find scanline crossings for segments
    # 0-2 and 1-2.  This loop is skipped if y1=y2.
    sa = dx12 * (y - y1)
    sb = dx02 * (y - y0)

    for y in range(y, y2 + 1):
        a = x1 + sa / dy12
        b = x0 + sb / dy02
        sa += dx12
        sb += dx02

        if a > b:
            a, b = b, a
            _draw_fast_hline(setter, a, y, b - a + 1, color, aa)


def draw_char(fonts, setter, width, height, x, y, c, color, bg,
              aa=False, font=font.default_font, font_scale=1):
    if font_scale < 1:
        log.error('font_scale %s must be >= 1', font_scale)
        font_scale = 1
    f = fonts[font]
    fh = f['height']
    FONT = f['data']

    c = ord(c)  # make it the int value
    if c < f['bounds'][0] or c > f['bounds'][1]:
        c_data = f['undef']
    else:
        c_data = FONT[c - f['bounds'][0]]

    fw = len(c_data)
    for i in range(fw + f['sep']):
        x_pos = x + (i * font_scale)
        if max(0, 1 - font_scale * fw) <= x_pos < width:
            if i >= fw:
                line = 0
            else:
                line = FONT[c][i]
            for j in range(fh):
                y_pos = y + (j * font_scale)
                if ((y_pos < height) and
                        (y_pos + fh * font_scale - 1) >= 0):
                    if line & 0x1:
                        if font_scale == 1:
                            setter(x_pos, y_pos, color)
                        else:
                            fill_rect(setter, x_pos, y_pos, font_scale,
                                      font_scale, color, aa)
                    elif bg != color and bg is not None:
                        if font_scale == 1:
                            setter(x_pos, y_pos, bg)
                        else:
                            fill_rect(setter, x_pos, y_pos, font_scale,
                                      font_scale, bg, aa)
                line >>= 1
    return fw + f['sep']


def draw_text(fonts, setter, text, width, height, x=0, y=0,
              color=None, bg=colors.Off, aa=False,
              font=font.default_font, font_scale=1):
    fh = fonts[font]['height']
    for c in text:
        if c == '\n':
            y += font_scale * fh
            x = 0
        elif c == '\r':
            pass  # skip it
        else:
            fw = draw_char(fonts, setter, width, height,
                           x, y, c, color, bg, aa, font, font_scale)
            x += font_scale * fw
            if x >= width:
                break



"""
Split is an animation.Collection that runs multiple smaller animations as
segments within a larger diplay.

"""

import copy
from . parallel import Parallel
from .. layout.strip import Strip


class Split(Parallel):

    def __init__(self, *args, size=100, **kwds):
        """
        Arguments --
            size: a number or a list of numbers representing the size of each
                segment from the original layout.  If there aren't enough sizes
                for each segment, the list of sizes is reused repeatedly.
        """
        super().__init__(*args, detach=False, **kwds)
        if not size:
            raise ValueError('Split must have at least one size')

        self.size = size if isinstance(size, list) else [size]
        self.is_numpy = hasattr(self.color_list, 'dtype')

        for animation, begin, end in self._foreach():
            animation.layout = Strip([], color_list=self.color_list[begin:end])

    def step(self, amt=1):
        super().step(amt)

        # If color_lists are numpy, then changing the sublists will
        # change the master list, so you don't have to do anything!
        #
        # Otherwise, copy the results into the classic Python color_list.
        if not self.is_numpy:
            for animation, begin, end in self._foreach():
                self.color_list[begin:end] = animation.color_list

    def _foreach(self):
        begin = 0
        for i, animation in enumerate(self.animations):
            end = begin + self.size[i % len(self.size)]
            if animation:
                yield animation, begin, end
            begin = end



import copy, multiprocessing
from . import opener, server, trigger_process
from .. animation import STATE
from .. import wrapper
from ... project import load
from ... util import log
from bibliopixel.animation.off import Off

DEFAULT_OFF = 'OFF_ANIM'
DEFAULT_ANIM_CONFIG = {
    'bgcolor': '#00ff00',
    'font_color': '#ffffff',
    'valid': True
}
DEFAULT_AUTO_DEMO_TIME = 10
DEFAULT_AUTO_DEMO_NAME = 'DEMO_ANIM'
OPENER_TIMEOUT = 2

BAD_DEFAULT_ERROR = """\
`{}` is not a valid default!
It must be one of the configured animation names."""


def normalize_name(name):
    return ''.join('_' if e is ' ' else e
                   for e in name
                   if e.isalnum() or e is ' ' or e is '_')


class RemoteControl(wrapper.Indexed):
    @staticmethod
    def pre_recursion(desc):
        wrapper.Indexed.pre_recursion(desc)

        animations = desc['animations']
        auto_demo = desc.pop('auto_demo', None)

        if auto_demo:
            auto_demo.setdefault('name', DEFAULT_AUTO_DEMO_NAME)
            animations.insert(0, auto_demo)

        desc['name_map'] = {}
        for i, anim in enumerate(animations):
            anim.setdefault('run', {}).update(threaded=True)

            display_name = anim.get('name') or str(i)

            anim['data'] = dict(DEFAULT_ANIM_CONFIG, **anim.get('data', {}))
            anim['data'].setdefault('display', display_name)

            # Get the normalized name - with only URL-safe characters in it.
            name = normalize_name(display_name)
            # It's a defect that we have name in two places but...?
            anim['name'] = anim['data']['name'] = name
            desc['name_map'][name] = i

        if len(desc['name_map']) < len(animations):
            log.warning('There are multiple animations with the same name. '
                        'Only the last will work.')

        if auto_demo:
            auto_demo.setdefault('typename', 'sequence')
            seconds = auto_demo['run'].pop('seconds', DEFAULT_AUTO_DEMO_TIME)
            auto_demo.setdefault('length', seconds)

            auto_demo['animations'] = [copy.deepcopy(a) for a in animations[1:]]
            for a in auto_demo['animations']:
                a['run']['threaded'] = False

            desc.setdefault('default', auto_demo['name'])

        default = desc.get('default', None)
        if default is not None:
            default = normalize_name(default)
            index = desc['name_map'].get(default)
            if index is None:
                log.warning('Do not understand default "%s"', default)
                log.warning('Names are %s', ', '.join(desc['name_map']))
            desc['default'] = index

        return desc

    def __init__(self, *args, name_map, external_access=False, port=5000,
                 title='BiblioPixel Remote', bgcolor='black',
                 font_color='white', default=None,
                 triggers=[], open_page=False, **kwds):
        super().__init__(*args, **kwds)
        self.internal_delay = 0  # never wait

        self.anim_cfgs = [a.data for a in self.animations]
        self.name_map = name_map
        self.port = port
        if open_page is False:
            self.open_page = False
        elif open_page is True:
            self.open_page = OPENER_TIMEOUT
        else:
            self.open_page = float(open_page)

        for anim in self.animations:
            anim.on_completion = self.on_completion

            if getattr(anim, 'empty', False):
                # The animation failed to load.
                display_name = anim.data.get('display', '(no error)')
                anim.data.update(
                    valid=False,
                    display='FAILED: ' + display_name,
                    bgcolor='rgb(48, 48, 48)',
                    font_color='white')

        if default is None:
            off = Off(self.layout)
            off._set_runner({'threaded': True})
            self.name_map[DEFAULT_OFF] = len(self.animations)
            self.animations.append(off)
        else:
            self.name_map[DEFAULT_OFF] = default

        self.index = self.name_map[DEFAULT_OFF]  # start with default animation

        self.ui_config = {
            'bgcolor': bgcolor,
            'font_color': font_color,
            'title': title
        }

        self.q_recv = multiprocessing.Queue()

        self.send_queues = {
            'RemoteServer': multiprocessing.Queue()
        }

        server_args = (
            external_access,
            port,
            self.send_queues['RemoteServer'],
            self.q_recv
        )

        self.server = multiprocessing.Process(target=server.run_server,
                                              args=server_args)
        if self.open_page is not False:
            opener.opener('localhost', port, self.open_page)

        self.handlers = {
            'run_animation': self.run_animation,
            'stop_animation': self.stop_animation,
            'get_config': self.get_config,
            'trigger_animation': self.run_animation,
            'brightness': self.change_brightness
        }

        self.trigger_procs = {}
        self.triggers = {}
        for trigger in triggers:
            typename = trigger.pop('typename')
            events = trigger.pop('events')
            if typename and events:
                load.code(typename)  # attempt early to fail early
                self.triggers.setdefault(typename, []).extend((events, trigger))
            else:
                raise ValueError(
                    'Triggers require `typename` and `events` fields!')

        for typename, trigger in self.triggers.items():
            events, kwargs = trigger
            self.trigger_procs[typename] = multiprocessing.Process(
                target=trigger_process.run_trigger,
                args=(typename, self.q_recv, events, kwargs))

    def cleanup(self, clean_layout=True):
        self.q_recv.close()
        for q in self.send_queues.values():
            q.close()
        self.server.terminate()
        for _, t in self.trigger_procs.items():
            t.terminate()
        super().cleanup(clean_layout)

    def on_completion(self, reason):
        if reason != STATE.canceled:
            self.stop_animation()

    # API Handlers
    def run_animation(self, name=DEFAULT_OFF):
        log.debug('run_animation %s', name)

        if name not in self.name_map:
            error = 'Invalid animation name: {}'.format(name)
            log.info(error)
            return False, error

        if self.current_animation:
            self.current_animation.cleanup(clean_layout=False)
            self.index = -1

        self.index = self.name_map[name]
        log.info('Running animation: {}'.format(name))
        self.current_animation.start()
        return True, None

    def stop_animation(self, data):
        return self.run_animation()

    def change_brightness(self, data):
        if isinstance(data, str):
            try:
                data = int(data)
            except:
                return False, 'Invalid brightness value!'

        self.layout.set_brightness(data)
        return True, None

    def get_config(self, data):
        resp = {
            'ui': self.ui_config,
            'brightness': self.layout.brightness,
            'animations': self.anim_cfgs,
        }
        return True, resp

    def pre_run(self):
        self.server.start()
        for proc in self.trigger_procs.values():
            proc.start()

    def step(self, amt=None):
        if not self.cur_step:
            self.run_animation()
        super().step(amt)
        while not self.threading.stop_event.isSet():
            recv = self.q_recv.get()
            req = recv['req']
            if req not in self.handlers:
                resp = False, '{} is not a valid request!'.format(req)
            else:
                resp = self.handlers[req](recv['data'])

            resp_q = self.send_queues.get(recv['sender'])
            if resp_q:
                resp_q.put(resp)



from . indexed import Indexed


class Wrapper(Indexed):
    # TODO: No unit tests cover any of this.
    @staticmethod
    def pre_recursion(desc):
        if 'animations' in desc:
            raise ValueError('Cannot specify animations in a Wrapper')
        desc['animations'] = [desc.pop('animation')]
        return Indexed.pre_recursion(desc)

    @property
    def animation(self):
        return self.animations[0]



from . animation import Animation
from .. util import log
import threading
from enum import IntEnum


class HOST_TYPE(IntEnum):
    STRIP = 1
    MATRIX = 2
    CIRCLE = 3
    CUBE = 4


HOST_MAP = {
    'Strip': HOST_TYPE.STRIP,
    'Matrix': HOST_TYPE.MATRIX,
    'Circle': HOST_TYPE.CIRCLE,
    'Cube': HOST_TYPE.CUBE,
}


class Receiver(Animation):
    free_run = True

    def __init__(self, layout, **kwds):
        super().__init__(layout, **kwds)
        name = type(self.layout).__name__

        if name not in HOST_MAP:
            raise ValueError('layout must be of type' + ', '.join(HOST_MAP))

        self.host_type = HOST_MAP[name]
        self._hold_for_data = threading.Event()
        self._stop_event = threading.Event()
        self._stop_event.clear()
        self._recv_thread_obj = None

    def pre_run(self):
        self.start()

    def start(self):
        self._t = threading.Thread(target=self._recv_thread_obj)
        self._t.setDaemon(True)  # don't hang on exit
        self._t.start()
        log.info("Receiver Listening on %s", self.address)

    def thread_cleanup(self):
        # To be overriden, if need be
        pass

    def stop(self):
        self._stop_event.set()
        log.info("Stopping Receiver...")
        self.thread_cleanup()

    def _exit(self, type, value, traceback):
        self.stop()

    def step(self, amt=1):
        """
        This may seem silly, but on a Receiver step() need not do anything.
        Instead, receive the data on the receive thread and set it on the buffer
        then call self._hold_for_data.set()
        """
        if not self._stop_event.isSet():
            self._hold_for_data.wait()
            self._hold_for_data.clear()


from .. util import deprecated
if deprecated.allowed():  # pragma: no cover
    BaseReceiver = Receiver



import flask, werkzeug.serving
from werkzeug.datastructures import ImmutableOrderedMultiDict

from ... util import log
from ... util.threads import runnable
from ... animation.remote import opener


class OrderedFlask(flask.Flask):
    # http://flask.pocoo.org/docs/1.0/patterns/subclassing/

    class request_class(flask.Request):
        parameter_storage_class = ImmutableOrderedMultiDict


class Server(runnable.LoopThread):
    OPEN_DELAY = 1

    def __init__(self, port, external_access, open_page, **kwds):
        super().__init__()
        self.port = port
        self.hostname = '0.0.0.0' if external_access else 'localhost'
        self.app = OrderedFlask(__name__, **kwds)
        self.open_page = open_page

    def run_once(self):
        if self.open_page:
            opener.raw_opener('localhost', self.port, self.OPEN_DELAY)
        werkzeug.serving.run_simple(self.hostname, self.port, self.app)
        super().stop()

    def stop(self):
        def error():
            log.error('Unable to shut down REST server on port %d', self.port)

        super().stop()
        try:
            flask.request.environ.get('werkzeug.server.shutdown', error)()
        except Exception as e:
            log.debug('Exception shutting werkzeug down')



import os, sys

CHOICES = 'ignore', 'fail', 'warn', 'warn_once'
DEFAULT = 'warn_once'
ACTION = None

HELP = """
Specify what to do when a project uses deprecated features:

  ignore: do nothing
  warn: print warning messages for each feature
  warn_once: print a warning message, but only once for each type of feature
  fail: throw an exception
"""

DEPRECATED = set()
FLAG = '--deprecated'
V4_FLAG = '--v4'
ENVIRONMENT_VARIABLE = 'BP_DEPRECATED'


def allowed():
    _compute_action()
    return ACTION != 'fail'


def deprecated(msg, *args, **kwds):
    _compute_action()

    if ACTION == 'ignore':
        return
    if ACTION == 'warn_once' and msg in DEPRECATED:
        return

    formatted = msg.format(*args, **kwds)
    if ACTION == 'fail':
        raise ValueError(formatted)

    DEPRECATED.add(msg)

    from . import log
    log.warning(formatted)


def _compute_action():
    global ACTION
    if ACTION:
        return

    if FLAG in sys.argv:
        raise ValueError('%s needs an argument (one of %s)' %
                         (FLAG, ', '.join(CHOICES)))

    if V4_FLAG in sys.argv:
        ACTION = 'fail'

    d = [i for i, v in enumerate(sys.argv) if v.startswith(FLAG + '=')]
    if len(d) > 1:
        raise ValueError('Only one %s argument can be used' % FLAG)

    if not d:
        ACTION = os.getenv(ENVIRONMENT_VARIABLE, ACTION or DEFAULT)

    else:
        arg = sys.argv.pop(d[0])
        _, *rest = arg.split('=')

        if len(rest) > 1:
            raise ValueError('Extra = in flag %s' % arg)

        if not (rest and rest[0].strip()):
            raise ValueError('%s needs an argument (one of %s)' %
                             (FLAG, ', '.join(CHOICES)))
        ACTION = rest[0]

    if ACTION not in CHOICES:
        ACTION = None
        raise ValueError('Unknown deprecation value (must be one of %s)' %
                         ', '.join(CHOICES))



import logging, sys
from logging import DEBUG, INFO, WARNING, ERROR

FRAME = DEBUG - 5
LOG_NAMES = {'frame': FRAME, 'debug': DEBUG, 'info': INFO, 'warning': WARNING,
             'error': ERROR}

SORTED_NAMES = tuple(k for k, v in sorted(LOG_NAMES.items()))


# From https://stackoverflow.com/a/35804945/43839
def _addLoggingLevel(levelName, levelNum, methodName=None):
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example
    -------
    >>> addLoggingLevel('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5

    """
    if not methodName:
        methodName = levelName.lower()

    if hasattr(logging, levelName):
        raise AttributeError(
            '{} already defined in logging module'.format(levelName))
    if hasattr(logging, methodName):
        raise AttributeError(
            '{} already defined in logging module'.format(methodName))
    if hasattr(logging.getLoggerClass(), methodName):
        raise AttributeError(
            '{} already defined in logger class'.format(methodName))

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def logForLevel(self, message, *args, **kwargs):
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)

    def logToRoot(message, *args, **kwargs):
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)


def _new_custom_logger(name='BiblioPixel',
                       fmt='%(levelname)s - %(module)s - %(message)s'):
    logger = logging.getLogger(name)
    formatter = logging.Formatter(fmt=fmt)

    def add_handler(level, outfile):
        class Filter(logging.Filter):
            def filter(self, rec):
                return rec.levelno == level

        h = logging.StreamHandler(outfile)
        h.setLevel(level)
        h.addFilter(Filter())
        h.setFormatter(formatter)
        logger.addHandler(h)

    if not logger.handlers:
        logger.setLevel(INFO)
        add_handler(FRAME, sys.stdout)
        add_handler(DEBUG, sys.stdout)
        add_handler(INFO, sys.stdout)
        add_handler(WARNING, sys.stderr)
        add_handler(ERROR, sys.stderr)

    else:  # pragma: no cover
        pass

    return logger


def set_log_level(level):
    """
    :param level: the level to set - either a string level name from
                  'frame', 'debug', 'info', 'warning', 'error'
                  or an integer log level from:
                  log.FRAME, log.DEBUG, log.INFO, log.WARNING, log.ERROR
    """
    if isinstance(level, str):
        level = LOG_NAMES[level.lower()]

    logger.setLevel(level)


def get_log_level():
    return logger.getEffectiveLevel()


def is_debug():
    return get_log_level() <= DEBUG


# Add a new logging level FRAME for messages which appear on every frame.
_addLoggingLevel('FRAME', FRAME)
logger = _new_custom_logger()

frame, debug, info, warning, error = (
    logger.frame, logger.debug, logger.info, logger.warning, logger.error)


# The function `printer` emits text no matter what the loglevel, and without any
# introducers like "INFO".  By default this is the same as the global `print` -
# re-assign this variable if you need to redirect your printing.
printer = print  # noqa: T001, T002


from . import deprecated

if deprecated.allowed():  # pragma: no cover
    def setLogLevel(level):
        deprecated.deprecated('util.setLogLevel')
        set_log_level(level)



"""
Toggle between color names and color tuples
"""

from .. util import colors, log


def run(args):
    if args.list:
        log.printer(*sorted(colors.COLOR_DICT.items()), sep='\n')

    if not args.colors:
        if not args.list:
            raise ValueError('No colors supplied!')

    failures = []
    for c in args.colors:
        try:
            log.printer(c, colors.toggle(c), sep=': ')
        except Exception as e:
            failures.append(c)
    if failures:
        s = '' if len(failures) == 1 else 's'
        failures = ', '.join('"%s"' % f for f in failures)
        raise ValueError('Did not understand color%s %s' % (s, failures))


def set_parser(parser):
    parser.set_defaults(run=run)
    parser.add_argument(
        'colors', nargs='*',
        help='Color names or tuples',
        default='')

    parser.add_argument(
        '-l', '--list', action='store_true',
        help='List all the colors')



from . strip import BaseStripAnim
from . matrix import BaseMatrixAnim
from . animation import Animation
from .. util import colors

BASE_COLORS = [colors.Red, colors.Green, colors.Green,
               colors.Blue, colors.Blue, colors.Blue]
CYCLE_COLORS = [colors.Red, colors.Green, colors.Blue, colors.White]


class StripChannelTest(BaseStripAnim):

    def __init__(self, layout):
        super().__init__(layout)
        self.internal_delay = 0.500
        self.colors = CYCLE_COLORS

    def step(self, amt=1):
        for i, c in enumerate(BASE_COLORS):
            self.layout.set(i, c)

        color = self.cur_step % 4
        self.layout.fill(self.colors[color], 7, 9)


class MatrixChannelTest(BaseMatrixAnim):

    def __init__(self, layout):
        super().__init__(layout, 0, 0)
        self.internal_delay = 0.500
        self.colors = [colors.Red, colors.Green, colors.Blue, colors.White]

    def step(self, amt=1):
        self.layout.drawLine(0, 0, 0, self.height - 1, colors.Red)
        self.layout.drawLine(1, 0, 1, self.height - 1, colors.Green)
        self.layout.drawLine(2, 0, 2, self.height - 1, colors.Green)
        self.layout.drawLine(3, 0, 3, self.height - 1, colors.Blue)
        self.layout.drawLine(4, 0, 4, self.height - 1, colors.Blue)
        self.layout.drawLine(5, 0, 5, self.height - 1, colors.Blue)

        color = self.cur_step % 4
        self.layout.fillRect(7, 0, 3, self.height, self.colors[color])


class MatrixCalibrationTest(BaseMatrixAnim):

    def __init__(self, layout):
        super().__init__(layout, 0, 0)
        self.internal_delay = 0.500
        self.colors = [colors.Red, colors.Green, colors.Green,
                       colors.Blue, colors.Blue, colors.Blue]

    def step(self, amt=1):
        self.layout.all_off()
        i = self.cur_step % self.width
        for x in range(i + 1):
            c = self.colors[x % len(self.colors)]
            self.layout.drawLine(x, 0, x, i, c)

        self.completed = (i == (self.width - 1))


class PixelTester(Animation):
    """"""
    PAUSE = 10

    def __init__(self, *args, brightness=1.0, **kwds):
        self.brightness = brightness
        super().__init__(*args, **kwds)

    def pre_run(self):
        self.stepper = self.steps()

    def step(self, amt=1):
        try:
            next(self.stepper)
        except StopIteration:
            self.completed = True

    def steps(self):
        for color in (colors.Red, colors.Green, colors.Blue, colors.Yellow,
                      colors.Fuchsia, colors.Aqua, colors.White):
            color = tuple(c * self.brightness for c in color)
            for i in range(len(self.color_list)):
                self.color_list[i] = color
                yield

            for i in range(self.PAUSE):
                yield

            self.color_list = [colors.Black] * len(self.color_list)
            yield



import io, os, pathlib
from .. import log


def write_gif(filename, frames, **gif_options):
    """
    Write a series of frames as a single animated GIF.

    :param str filename: the name of the GIF file to write

    :param list frames: a list of filenames, each of which represents a single
        frame of the animation.  Each frame must have exactly the same
        dimensions, and the code has only been tested with .gif files.

    The whole gif_options dictionary is passed as keywords to PIL.Image.save.
    Its individual options are documented here as taken from the PIL
    documentation:

    :param int loop:
        The number of iterations. Default 0 (meaning loop indefinitely).

    :param float, list duration:
        The duration (in seconds) of each frame. Either specify one value
        that is used for all frames, or one value for each frame.
        Note that in the GIF format the duration/delay is expressed in
        hundredths of a second, which limits the precision of the duration.

    :param float fps:
        The number of frames per second. If duration is not given, the
        duration for each frame is set to 1/fps. Default 10.

    :param int palette:
        The number of colors to quantize the image to. Is rounded to
        the nearest power of two. Default 256.
    """

    from PIL import Image
    images = []
    for f in frames:
        data = open(f, 'rb').read()
        images.append(Image.open(io.BytesIO(data)))

    im = images.pop(0)

    if 'loop' not in gif_options:
        gif_options['loop'] = 0
        # See https://github.com/python-pillow/Pillow/issues/3255

    im.save(filename, save_all=True, append_images=images, **gif_options)


"""
This is probably not used and should likely be replaced by the code
in BiblioPixelAnimations.matrix.ImageAnim
"""


from .. import deprecated
if deprecated.allowed():  # pragma: no cover
    def convert_mode(image, mode='RGB'):
        """Return an image in the given mode."""
        deprecated.deprecated('util.gif.convert_model')

        return image if (image.mode == mode) else image.convert(mode=mode)

    def image_to_colorlist(image, container=list):
        """Given a PIL.Image, returns a ColorList of its pixels."""
        deprecated.deprecated('util.gif.image_to_colorlist')

        return container(convert_mode(image).getdata())

    def animated_gif_to_colorlists(image, container=list):
        """
        Given an animated GIF, return a list with a colorlist for each frame.
        """
        deprecated.deprecated('util.gif.animated_gif_to_colorlists')

        from PIL import ImageSequence

        it = ImageSequence.Iterator(image)
        return [image_to_colorlist(i, container) for i in it]


if __name__ == '__main__':
    pass



from ... util.colors import gamma
from .. channel_order import ChannelOrder
from . base import SPIBase


class LPD8806(SPIBase):
    """Main driver for LPD8806 based LED strips on devices like the Raspberry Pi
       and BeagleBone.

    Provides the same parameters as
    :py:class:`bibliopixel.drivers.SPI.SPIBase`
    """

    def __init__(self, num, gamma=gamma.LPD8806, **kwargs):
        super().__init__(num, gamma=gamma, **kwargs)

        # LPD8806 requires latch bytes at the end
        self._latchBytes = (self.numLEDs + 31) // 32
        for i in range(0, self._latchBytes):
            self._buf.append(0)

    # LPD8806 requires gamma correction and only supports 7-bits per channel
    # running each value through gamma will fix all of this.



from . ops import Ops
from . editor import Editor
from . receiver import Receiver


class Action(Receiver):
    """
    An Action takes an incoming message, applies Ops to it, and then
    uses it to set a value on a Editor.
    """

    def __init__(self, address, ops=()):
        self.address = Editor(address)
        self.ops = Ops(*ops)

    def set_root(self, root):
        self.address.set_root(root)

    def receive(self, values):
        if self.ops:
            if len(values) == 1:
                values = [self.ops(values[0])]
            else:
                # TODO: They specified ops, but we can't use it.
                # Should we warn here?  Can we use the ops somehow?
                pass
        return self.address.receive(values)

    def __bool__(self):
        return bool(self.address or self.ops)

    def __str__(self):
        if self.ops:
            return '%s->%s' % self.address, self.ops
        return str(self.address)

    @classmethod
    def make(cls, action):
        if isinstance(action, str):
            return cls(action)
        if isinstance(action, dict):
            return cls(**action)
        return cls(*action)


class ActionList(Receiver):
    """A list of Actions."""

    def __init__(self, actions=None):
        if isinstance(actions, (str, dict)):
            actions = [actions]
        self.actions = tuple(Action.make(a) for a in actions or ())

    def set_root(self, root):
        for a in self.actions:
            a.set_root(root)

    def receive(self, msg):
        values = tuple(msg.values())
        for a in self.actions:
            a.receive(values)

    def __bool__(self):
        return bool(self.actions)

    def __str__(self):
        return ' + '.join(str(a) for a in self.actions)



import threading, time
from .. import log
from . import compose_events, runnable


class UpdateDriverThread(runnable.LoopThread):

    def __init__(self, driver):
        super().__init__()
        self.setDaemon(True)
        self._wait = threading.Event()
        self._reading = threading.Event()
        self._reading.set()
        self._updating = threading.Event()
        self._driver = driver

    def update_colors(self):
        self._reading.wait()
        self._reading.clear()
        self._wait.set()

    def sync(self):
        self._updating.wait()
        self._driver.sync()

    def sending(self):
        return self._wait.isSet()

    def run_once(self):
        self._wait.wait()
        self._updating.clear()
        self._driver.update_colors()
        self._wait.clear()
        self._reading.set()
        self._updating.set()


class UpdateThread(runnable.LoopThread):

    def __init__(self, drivers):
        super().__init__()
        self.setDaemon(True)
        self._wait = threading.Event()
        self._reading = threading.Event()
        self._reading.set()
        self._drivers = drivers
        for d in self._drivers:
            t = UpdateDriverThread(d)
            t.start()
            d._thread = t

        events = (d._thread._updating for d in self._drivers)
        self._updated = compose_events.compose_events(events)

    def update_colors(self):
        self._reading.wait()
        for d in self._drivers:
            d._thread.update_colors()
        self._reading.clear()
        self._wait.set()

    def run_once(self):
        self._wait.wait()
        self._updated.wait()

        for d in self._drivers:
            d._thread.sync()

        self._updated.clear()
        self._wait.clear()
        self._reading.set()


class NoThreading(runnable.Runnable):
    """
    """

    def __init__(self, layout):
        super().__init__()
        self.layout = layout

    def update_colors(self):
        for d in self.layout.drivers:
            d.update_colors()
        for d in self.layout.drivers:
            d.sync()

    def wait_for_update(self):
        pass

    def push_to_driver(self):
        """Push the current pixel state to the driver"""
        self.wait_for_update()
        self.update_colors()


class UseThreading(NoThreading):
    def __init__(self, layout):
        self.layout = layout
        self.update_thread = UpdateThread(layout.drivers)
        self.update_thread.start()

    def update_colors(self):
        self.update_thread.update_colors()

    def wait_for_update(self):
        while all([d._thread.sending() for d in self.layout.drivers]):
            time.sleep(0.000001)

    def wait(self):
        self.update_thread.wait()

    def stop(self):
        self.update_thread.stop()


def UpdateThreading(enable, layout):
    """
    UpdateThreading handles threading - and eventually multiprocessing - for
    Layout.
    """
    return (UseThreading if enable else NoThreading)(layout)



from . import wrapper
from .. util import color_list
from .. util.limit import Limit


class Limiter(wrapper.Wrapper):
    def __init__(self, *args, limit=None, **kwds):
        """
        :param dict limit: A construction dictionary for a Limit.
        """
        super().__init__(*args, **kwds)
        self.limit = Limit(**(limit or {}))
        self._math = color_list.Math(self.color_list)

    def pre_run(self):
        super().pre_run()
        self.animation.layout = self.layout.clone()

    def step(self, amt=1):
        super().step(amt)
        self._math.copy(self.color_list, self.animation.color_list)
        self.limit.limit_colors(self.color_list, self._math)



import os, pathlib
from .. project.importer import import_module
from .. util import deprecated, log

COMMANDS = (
    'animations', 'all_pixel', 'all_pixel_test', 'clear_cache', 'color', 'demo',
    'devices', 'info', 'monitor', 'new', 'run')

if deprecated.allowed():  # pragma: no cover
    SIGNAL_COMMANDS = 'kill', 'pid', 'restart', 'shutdown'
    COMMANDS += SIGNAL_COMMANDS

MODULES = {c: import_module('bibliopixel.main.' + c) for c in COMMANDS}


BP_DELETED_COMMANDS = False

if os.environ.get('BP_DELETED_COMMANDS', BP_DELETED_COMMANDS):
    DELETED = (
        'list', 'load', 'remove', 'reset', 'save', 'set', 'show', 'update')
    MODULES.update({
        c: import_module('bibliopixel.main.deleted.' + c) for c in DELETED})

FILE = pathlib.Path(__file__).parent / 'commands.rst.tmpl'

HELP = FILE.open().read().format(
    command_count=len(COMMANDS),
    commands=[
        ', '.join(COMMANDS[0:8]),
        ', '.join(COMMANDS[8:])])

BP_HEADER = """
APPENDIX: ``bp <command> --help`` for each command
==================================================
"""

BP_TEMPLATE = """\
``bp {command}``
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

{doc}

{description}
"""

SEPARATOR = """
------------------------------------

"""


def get_command_help(command):
    module = MODULES[command]
    return BP_TEMPLATE.format(
        command=command,
        module=module,
        doc=module.__doc__.strip(),
        description=getattr(module, 'DESCRIPTION', ''))


BP_HELP = SEPARATOR.join(get_command_help(c) for c in COMMANDS)


def print_help():
    print(HELP)
    print(BP_HEADER)
    print(BP_HELP)


if __name__ == '__main__':
    print_help()



from . load_image import loadImage
from . reshape import crop, resize

from .. import deprecated
if deprecated.allowed():  # pragma: no cover
    from . load_image import show_image, showImage, loadImage
    from . gif import (
        animated_gif_to_colorlists, convert_mode, image_to_colorlist)



import getpass, platform, sys, threading
from .. util import log
from . control import ExtractedControl

# See https://stackoverflow.com/questions/42603000
DARWIN_ROOT_WARNING = """
In MacOS, pynput must to be running as root in order to get keystrokes.

Try running your program like this:

    sudo %s <your commands here>
"""

INSTALL_ERROR = """
Please install the pynput library with

    $ pip install pynput

"""


try:
    import pynput

except ImportError:
    pynput = Listener = None

else:
    class Listener(pynput.keyboard.Listener):
        def join(self, timeout=None):
            # join() on pynput.keyboard.Listener waits on a queue...
            self._queue.put(None)
            return super().join(timeout)


def keyname(key):
    return getattr(key, 'name', None) or getattr(key, 'char')


class Keyboard(ExtractedControl):
    EXTRACTOR = {
        'keys_by_type': {
            'press': ['type', 'key'],
            'release': ['type', 'key'],
        },

        'normalizers': {
            'key': keyname,
        },
    }

    def _press(self, key):
        self.receive({'type': 'press', 'key': key})

    def _release(self, key):
        self.receive({'type': 'release', 'key': key})

    def _make_thread(self):
        if not pynput:
            raise ValueError(INSTALL_ERROR)

        if platform.platform().startswith('Darwin'):
            if getpass.getuser() != 'root':
                log.warning(DARWIN_ROOT_WARNING, sys.argv[0])

        log.info('Starting to listen for keyboard input')
        return Listener(self._press, self._release)



from enum import IntEnum


class LEDTYPE(IntEnum):
    """Enumerated LED type names to be used with
    :py:mod:`bibliopixel.drivers.serial` and
    :py:mod:`bibliopixel.drivers.SPI`
    """
    GENERIC = 0  # Use if the serial device only supports one chipset
    LPD8806 = 1
    WS2801 = 2
    # These are all the same
    WS2811 = 3
    WS2812 = 3
    WS2812B = 3
    NEOPIXEL = 3
    APA104 = 3
    # 400khz variant of above
    WS2811_400 = 4

    TM1809 = 5
    TM1804 = 5
    TM1803 = 6
    UCS1903 = 7
    SM16716 = 8
    APA102 = 9
    SK9822 = 9
    LPD1886 = 10
    P9813 = 11



"""Functions to convert HSV colors to RGB colors lovingly ported from FastLED

The basically fall into two groups: spectra, and rainbows.

Spectra and rainbows are not the same thing.  Wikipedia has a good
illustration here

 http://upload.wikimedia.org/wikipedia/commons/f/f6/Prism_compare_rainbow_01.png

from this article

  http://en.wikipedia.org/wiki/Rainbow#Number_of_colours_in_spectrum_or_rainbow

that shows a 'spectrum' and a 'rainbow' side by side.  Among other
differences, you'll see that a 'rainbow' has much more yellow than
a plain spectrum.  "Classic" LED color washes are spectrum based, and
usually show very little yellow.

Wikipedia's page on HSV color space, with pseudocode for conversion
to RGB color space

  http://en.wikipedia.org/wiki/HSL_and_HSV

Note that their conversion algorithm, which is (naturally) very popular
is in the "maximum brightness at any given hue" style, vs the "uniform
brightness for all hues" style.

You can't have both; either purple is the same brightness as red, e.g

  red = #FF0000 and purple = #800080 -> same "total light" output

OR purple is 'as bright as it can be', e.g.

  red = #FF0000 and purple = #FF00FF -> purple is much brighter than red.

The colorspace conversions here try to keep the apparent brightness
constant even as the hue varies.

Adafruit's "Wheel" function, discussed here

  http://forums.adafruit.com/viewtopic.php?f=47&t=22483

is also of the "constant apparent brightness" variety.

More details here: https://github.com/FastLED/FastLED/wiki/FastLED-HSV-Colors
"""

from . arithmetic import color_blend, color_scale
from . conversions import hsv2rgb_spectrum, hsv2rgb_rainbow, hsv2rgb_360
from . conversions import hsv2rgb, hue2rgb, hue_gradient, hue2rgb_360
from . conversions import hue_helper, hue_helper360
from . wheel import wheel_color, wheel_helper
from . names import (
    COLORS, COLOR_DICT, color_to_name, name_to_color, to_color, toggle)

from .. import deprecated

if deprecated.allowed():  # pragma: no cover
    hex2rgb = COLORS.__getitem__
    from . import conversions as hue

    # Legacy color names
    from . classic import *  # noqa: F403



import os, shutil, statistics, tempfile, time, traceback
from . import gif, renderer
from .. import colors, log


DEFAULT_RENDER = {
    'color': colors.Black,
    'pixel_width': 2,
    'pixel_height': None,
    'ellipse': True,
    'vertical': False,
    'frame': 3,
    'padding': 1,
}


class GifWriter:
    """Write an animated GIF given frames from an animation."""

    def __init__(self, filename='gif_writer', render=None,
                 divide=1, frames=128, time=10, scale=1.0, gif_options=None,
                 tmp_dir=None):
        self.render = dict(DEFAULT_RENDER, **(render or {}))
        self.divide = divide
        self.frames = frames
        self.time = time
        self.frame_files = []
        self.scale = scale
        self.gif_options = gif_options or {}

        filename = os.path.expanduser(os.path.abspath(filename))
        file_root = filename[:-4] if filename.endswith('.gif') else filename
        self.filename = file_root + '.gif'
        os.makedirs(os.path.dirname(self.filename), exist_ok=True)

        if tmp_dir:
            self.tmp_dir_name = tmp_dir
            shutil.rmtree(tmp_dir, ignore_errors=True)
            os.makedirs(self.tmp_dir_name, exist_ok=True)
        else:
            self.tmp_dir = tempfile.TemporaryDirectory()
            self.tmp_dir_name = self.tmp_dir.name
        self.basename = os.path.basename(file_root)
        self.finished = False
        self.stop_after_write = True
        self.times = []

    def set_project(self, project):
        self.project = project
        self.render = renderer.renderer(project.layout, **self.render)
        assert self.render

    def step(self, cur_step):
        if self.finished or (self.divide >= 1 and cur_step % self.divide):
            return True

        frame = cur_step / max(self.divide, 1)
        self.times.append(self.project.time())

        if self.time:
            elapsed = self.times[-1] - self.times[0]
            self.finished = (elapsed >= self.time)
        else:
            self.finished = (frame >= self.frames)

        if self.finished:
            self.write()
            return True

        frame_name = '%s%04d.gif' % (self.basename, frame)
        filename = os.path.join(self.tmp_dir_name, frame_name)

        self.render().save(filename)
        self.frame_files.append(filename)

    def write(self):
        if not self.frame_files:
            return

        ff, self.frame_files = self.frame_files, []
        go = dict(self.gif_options)

        if 'duration' in go:
            go.pop('fps', None)
            go['duration'] *= self.scale
            assert isinstance(go['duration'], (int, float))

        elif 'fps' in go:
            go['fps'] /= self.scale

        else:
            # Compute the average duration
            dt = [t1 - t0 for t0, t1 in zip(self.times, self.times[1:])]
            duration = statistics.mean(dt)

            # GIF durations are only expressed in hundredths of seconds.
            duration = round(duration, 2)
            go = dict(go, duration=duration)

        gif.write_gif(self.filename, ff, **go)
        self.tmp_dir = None



import functools
from ...drivers.channel_order import ChannelOrder

USAGE = """
A ChannelOrder can be initialized with:

* A list, tuple or string of lengths three: '012', 'bgr', (0, 2, 1)
* An integer from 0 to 5."""

NAMES = {
    'r': 0,
    'g': 1,
    'b': 2,
    'R': 0,
    'G': 1,
    'B': 2,
    '0': 0,
    '1': 1,
    '2': 2,
    0: 0,
    1: 1,
    2: 2,
}


@functools.singledispatch
def make(c):
    raise ValueError("Don't understand type %s" % type(c), USAGE)


@make.register(tuple)
@make.register(list)
@make.register(str)
def _(c):
    i = ChannelOrder.ORDERS.index(tuple(NAMES[i] for i in c))
    return ChannelOrder.ORDERS[i]


@make.register(int)
def _(c):
    if c < 0:
        raise IndexError('RGB indices cannot be negative')
    return ChannelOrder.ORDERS[c]



"""
A longer list of color names.

These colors come from:
https://github.com/timedata-org/timedata/blob
/master/src/cpp/timedata/color/names_table_inl.h

which came from Juce: http://juce.com.
"""

COLORS = {
    'alice blue': 0xf0f8ff,
    'antique white 1': 0xffefdb,
    'antique white 2': 0xeedfcc,
    'antique white 3': 0xcdc0b0,
    'antique white 4': 0x8b8378,
    'antique white': 0xfaebd7,
    'aqua': 0x00ffff,
    'aquamarine 1': 0x7fffd4,
    'aquamarine 2': 0x76eec6,
    'aquamarine 3': 0x66cdaa,
    'aquamarine 4': 0x458b74,
    'aquamarine': 0x7fffd4,
    'azure 1': 0xf0ffff,
    'azure 2': 0xe0eeee,
    'azure 3': 0xc1cdcd,
    'azure 4': 0x838b8b,
    'azure': 0xf0ffff,
    'banana': 0xe3cf57,
    'beige': 0xf5f5dc,
    'bisque 1': 0xffe4c4,
    'bisque 2': 0xeed5b7,
    'bisque 3': 0xcdb79e,
    'bisque 4': 0x8b7d6b,
    'bisque': 0xffe4c4,
    'black': 0x000000,
    'blanched almond': 0xffebcd,
    'blue 2': 0x0000ee,
    'blue 3': 0x0000cd,
    'blue 4': 0x00008b,
    'blue violet': 0x8a2be2,
    'blue': 0x0000ff,
    'brick': 0x9c661f,
    'brown 1': 0xff4040,
    'brown 2': 0xee3b3b,
    'brown 3': 0xcd3333,
    'brown 4': 0x8b2323,
    'brown': 0xa52a2a,
    'burly wood 1': 0xffd39b,
    'burly wood 2': 0xeec591,
    'burly wood 3': 0xcdaa7d,
    'burly wood 4': 0x8b7355,
    'burly wood': 0xdeb887,
    'burnt sienna': 0x8a360f,
    'burnt umber': 0x8a3324,
    'cadet blue 1': 0x98f5ff,
    'cadet blue 2': 0x8ee5ee,
    'cadet blue 3': 0x7ac5cd,
    'cadet blue 4': 0x53868b,
    'cadet blue': 0x5f9ea0,
    'cadmium orange': 0xff6103,
    'cadmium yellow': 0xff9912,
    'carrot': 0xed9121,
    'chartreuse 1': 0x7fff00,
    'chartreuse 2': 0x76ee00,
    'chartreuse 3': 0x66cd00,
    'chartreuse 4': 0x458b00,
    'chartreuse': 0x7fff00,
    'chocolate 1': 0xff7f24,
    'chocolate 2': 0xee7621,
    'chocolate 3': 0xcd661d,
    'chocolate 4': 0x8b4513,
    'chocolate': 0xd2691e,
    'cobalt green': 0x3d9140,
    'cobalt': 0x3d59ab,
    'cold grey': 0x808a87,
    'coral 1': 0xff7256,
    'coral 2': 0xee6a50,
    'coral 3': 0xcd5b45,
    'coral 4': 0x8b3e2f,
    'coral': 0xff7f50,
    'corn silk 1': 0xfff8dc,
    'corn silk 2': 0xeee8cd,
    'corn silk 3': 0xcdc8b1,
    'corn silk 4': 0x8b8878,
    'corn silk': 0xfff8dc,
    'cornflower blue': 0x6495ed,
    'crimson': 0xdc143c,
    'cyan 2': 0x00eeee,
    'cyan 3': 0x00cdcd,
    'cyan 4': 0x008b8b,
    'cyan': 0x00ffff,
    'dark blue': 0x00008b,
    'dark cyan': 0x008b8b,
    'dark goldenrod 1': 0xffb90f,
    'dark goldenrod 2': 0xeead0e,
    'dark goldenrod 3': 0xcd950c,
    'dark goldenrod 4': 0x8b6508,
    'dark goldenrod': 0xb8860b,
    'dark green': 0x006400,
    'dark grey': 0x555555,
    'dark khaki': 0xbdb76b,
    'dark magenta': 0x8b008b,
    'dark olive green': 0x556b2f,
    'dark olivegreen 1': 0xcaff70,
    'dark olivegreen 2': 0xbcee68,
    'dark olivegreen 3': 0xa2cd5a,
    'dark olivegreen 4': 0x6e8b3d,
    'dark olivegreen': 0x556b2f,
    'dark orange 1': 0xff7f00,
    'dark orange 2': 0xee7600,
    'dark orange 3': 0xcd6600,
    'dark orange 4': 0x8b4500,
    'dark orange': 0xff8c00,
    'dark orchid 1': 0xbf3eff,
    'dark orchid 2': 0xb23aee,
    'dark orchid 3': 0x9a32cd,
    'dark orchid 4': 0x68228b,
    'dark orchid': 0x9932cc,
    'dark red': 0x8b0000,
    'dark salmon': 0xe9967a,
    'dark sea green 1': 0xc1ffc1,
    'dark sea green 2': 0xb4eeb4,
    'dark sea green 3': 0x9bcd9b,
    'dark sea green 4': 0x698b69,
    'dark sea green': 0x8fbc8f,
    'dark slate blue': 0x483d8b,
    'dark slate grey 1': 0x97ffff,
    'dark slate grey 2': 0x8deeee,
    'dark slate grey 3': 0x79cdcd,
    'dark slate grey 4': 0x528b8b,
    'dark slate grey': 0x2f4f4f,
    'dark turquoise': 0x00ced1,
    'dark violet': 0x9400d3,
    'deep pink 1': 0xff1493,
    'deep pink 2': 0xee1289,
    'deep pink 3': 0xcd1076,
    'deep pink 4': 0x8b0a50,
    'deep pink': 0xff1493,
    'deep sky blue 1': 0x00bfff,
    'deep sky blue 2': 0x00b2ee,
    'deep sky blue 3': 0x009acd,
    'deep sky blue 4': 0x00688b,
    'deep sky blue': 0x00bfff,
    'dim grey': 0x696969,
    'dodger blue 1': 0x1e90ff,
    'dodger blue 2': 0x1c86ee,
    'dodger blue 3': 0x1874cd,
    'dodger blue 4': 0x104e8b,
    'dodger blue': 0x1e90ff,
    'eggshell': 0xfce6c9,
    'emerald green': 0x00c957,
    'fire brick 1': 0xff3030,
    'fire brick 2': 0xee2c2c,
    'fire brick 3': 0xcd2626,
    'fire brick 4': 0x8b1a1a,
    'fire brick': 0xb22222,
    'flesh': 0xff7d40,
    'floral white': 0xfffaf0,
    'forest green': 0x228b22,
    'fuchsia': 0xff00ff,
    'gainsboro': 0xdcdcdc,
    'ghost white': 0xf8f8ff,
    'gold 1': 0xffd700,
    'gold 2': 0xeec900,
    'gold 3': 0xcdad00,
    'gold 4': 0x8b7500,
    'gold': 0xffd700,
    'goldenrod 1': 0xffc125,
    'goldenrod 2': 0xeeb422,
    'goldenrod 3': 0xcd9b1d,
    'goldenrod 4': 0x8b6914,
    'goldenrod': 0xdaa520,
    'gray': 0x808080,
    'green 1': 0x00ff00,
    'green 2': 0x00ee00,
    'green 3': 0x00cd00,
    'green 4': 0x008b00,
    'green yellow': 0xadff2f,
    'green': 0x00ff00,
    'grey': 0x808080,
    'honeydew 1': 0xf0fff0,
    'honeydew 2': 0xe0eee0,
    'honeydew 3': 0xc1cdc1,
    'honeydew 4': 0x838b83,
    'honeydew': 0xf0fff0,
    'hot pink 1': 0xff6eb4,
    'hot pink 2': 0xee6aa7,
    'hot pink 3': 0xcd6090,
    'hot pink 4': 0x8b3a62,
    'hot pink': 0xff69b4,
    'indian red 1': 0xff6a6a,
    'indian red 2': 0xee6363,
    'indian red 3': 0xcd5555,
    'indian red 4': 0x8b3a3a,
    'indian red': 0xcd5c5c,
    'indigo': 0x4b0082,
    'ivory 1': 0xfffff0,
    'ivory 2': 0xeeeee0,
    'ivory 3': 0xcdcdc1,
    'ivory 4': 0x8b8b83,
    'ivory black': 0x292421,
    'ivory': 0xfffff0,
    'khaki 1': 0xfff68f,
    'khaki 2': 0xeee685,
    'khaki 3': 0xcdc673,
    'khaki 4': 0x8b864e,
    'khaki': 0xf0e68c,
    'lavender blush 1': 0xfff0f5,
    'lavender blush 2': 0xeee0e5,
    'lavender blush 3': 0xcdc1c5,
    'lavender blush 4': 0x8b8386,
    'lavender blush': 0xfff0f5,
    'lavender': 0xe6e6fa,
    'lawn green': 0x7cfc00,
    'lemon chiffon 1': 0xfffacd,
    'lemon chiffon 2': 0xeee9bf,
    'lemon chiffon 3': 0xcdc9a5,
    'lemon chiffon 4': 0x8b8970,
    'lemon chiffon': 0xfffacd,
    'light blue 1': 0xbfefff,
    'light blue 2': 0xb2dfee,
    'light blue 3': 0x9ac0cd,
    'light blue 4': 0x68838b,
    'light blue': 0xadd8e6,
    'light coral': 0xf08080,
    'light cyan 1': 0xe0ffff,
    'light cyan 2': 0xd1eeee,
    'light cyan 3': 0xb4cdcd,
    'light cyan 4': 0x7a8b8b,
    'light cyan': 0xe0ffff,
    'light goldenrod 1': 0xffec8b,
    'light goldenrod 2': 0xeedc82,
    'light goldenrod 3': 0xcdbe70,
    'light goldenrod 4': 0x8b814c,
    'light goldenrod yellow': 0xfafad2,
    'light green': 0x90ee90,
    'light grey': 0xd3d3d3,
    'light pink 1': 0xffaeb9,
    'light pink 2': 0xeea2ad,
    'light pink 3': 0xcd8c95,
    'light pink 4': 0x8b5f65,
    'light pink': 0xffb6c1,
    'light salmon 1': 0xffa07a,
    'light salmon 2': 0xee9572,
    'light salmon 3': 0xcd8162,
    'light salmon 4': 0x8b5742,
    'light salmon': 0xffa07a,
    'light sea green': 0x20b2aa,
    'light sky blue 1': 0xb0e2ff,
    'light sky blue 2': 0xa4d3ee,
    'light sky blue 3': 0x8db6cd,
    'light sky blue 4': 0x607b8b,
    'light sky blue': 0x87cefa,
    'light slate blue': 0x8470ff,
    'light slate grey': 0x778899,
    'light steel blue 1': 0xcae1ff,
    'light steel blue 2': 0xbcd2ee,
    'light steel blue 3': 0xa2b5cd,
    'light steel blue 4': 0x6e7b8b,
    'light steel blue': 0xb0c4de,
    'light yellow 1': 0xffffe0,
    'light yellow 2': 0xeeeed1,
    'light yellow 3': 0xcdcdb4,
    'light yellow 4': 0x8b8b7a,
    'light yellow': 0xffffe0,
    'lime green': 0x32cd32,
    'lime': 0x00ff00,
    'limegreen': 0x32cd32,
    'linen': 0xfaf0e6,
    'magenta 2': 0xee00ee,
    'magenta 3': 0xcd00cd,
    'magenta 4': 0x8b008b,
    'magenta': 0xff00ff,
    'manganese blue': 0x03a89e,
    'maroon 1': 0xff34b3,
    'maroon 2': 0xee30a7,
    'maroon 3': 0xcd2990,
    'maroon 4': 0x8b1c62,
    'maroon': 0x800000,
    'medium aquamarine': 0x66cdaa,
    'medium blue': 0x0000cd,
    'medium orchid 1': 0xe066ff,
    'medium orchid 2': 0xd15fee,
    'medium orchid 3': 0xb452cd,
    'medium orchid 4': 0x7a378b,
    'medium orchid': 0xba55d3,
    'medium purple 1': 0xab82ff,
    'medium purple 2': 0x9f79ee,
    'medium purple 3': 0x8968cd,
    'medium purple 4': 0x5d478b,
    'medium purple': 0x9370db,
    'medium sea green': 0x3cb371,
    'medium seagreen': 0x3cb371,
    'medium slate blue': 0x7b68ee,
    'medium slateblue': 0x7b68ee,
    'medium spring green': 0x00fa9a,
    'medium turquoise': 0x48d1cc,
    'medium violet red': 0xc71585,
    'medium violetred': 0xc71585,
    'melon': 0xe3a869,
    'midnight blue': 0x191970,
    'mint cream': 0xf5fffa,
    'mint': 0xbdfcc9,
    'misty rose 1': 0xffe4e1,
    'misty rose 2': 0xeed5d2,
    'misty rose 3': 0xcdb7b5,
    'misty rose 4': 0x8b7d7b,
    'misty rose': 0xffe4e1,
    'moccasin': 0xffe4b5,
    'navajo white 1': 0xffdead,
    'navajo white 2': 0xeecfa1,
    'navajo white 3': 0xcdb38b,
    'navajo white 4': 0x8b795e,
    'navajo white': 0xffdead,
    'navy': 0x000080,
    'none': 0x000000,
    'old lace': 0xfdf5e6,
    'olive drab 1': 0xc0ff3e,
    'olive drab 2': 0xb3ee3a,
    'olive drab 3': 0x9acd32,
    'olive drab 4': 0x698b22,
    'olive drab': 0x6b8e23,
    'olive': 0x808000,
    'orange 1': 0xffa500,
    'orange 2': 0xee9a00,
    'orange 3': 0xcd8500,
    'orange 4': 0x8b5a00,
    'orange red 1': 0xff4500,
    'orange red 2': 0xee4000,
    'orange red 3': 0xcd3700,
    'orange red 4': 0x8b2500,
    'orange red': 0xff4500,
    'orange': 0xffa500,
    'orchid 1': 0xff83fa,
    'orchid 2': 0xee7ae9,
    'orchid 3': 0xcd69c9,
    'orchid 4': 0x8b4789,
    'orchid': 0xda70d6,
    'pale goldenrod': 0xeee8aa,
    'pale green 1': 0x9aff9a,
    'pale green 2': 0x90ee90,
    'pale green 3': 0x7ccd7c,
    'pale green 4': 0x548b54,
    'pale green': 0x98fb98,
    'pale turquoise 1': 0xbbffff,
    'pale turquoise 2': 0xaeeeee,
    'pale turquoise 3': 0x96cdcd,
    'pale turquoise 4': 0x668b8b,
    'pale turquoise': 0xafeeee,
    'pale violet red 1': 0xff82ab,
    'pale violet red 2': 0xee799f,
    'pale violet red 3': 0xcd6889,
    'pale violet red 4': 0x8b475d,
    'pale violet red': 0xdb7093,
    'papaya whip': 0xffefd5,
    'peachpuff 1': 0xffdab9,
    'peachpuff 2': 0xeecbad,
    'peachpuff 3': 0xcdaf95,
    'peachpuff 4': 0x8b7765,
    'peachpuff': 0xffdab9,
    'peacock': 0x33a1c9,
    'peru': 0xcd853f,
    'pink 1': 0xffb5c5,
    'pink 2': 0xeea9b8,
    'pink 3': 0xcd919e,
    'pink 4': 0x8b636c,
    'pink': 0xffc0cb,
    'plum 1': 0xffbbff,
    'plum 2': 0xeeaeee,
    'plum 3': 0xcd96cd,
    'plum 4': 0x8b668b,
    'plum': 0xdda0dd,
    'powder blue': 0xb0e0e6,
    'purple 1': 0x9b30ff,
    'purple 2': 0x912cee,
    'purple 3': 0x7d26cd,
    'purple 4': 0x551a8b,
    'purple': 0x800080,
    'raspberry': 0x872657,
    'raw sienna': 0xc76114,
    'red 1': 0xff0000,
    'red 2': 0xee0000,
    'red 3': 0xcd0000,
    'red 4': 0x8b0000,
    'red': 0xff0000,
    'rosy brown 1': 0xffc1c1,
    'rosy brown 2': 0xeeb4b4,
    'rosy brown 3': 0xcd9b9b,
    'rosy brown 4': 0x8b6969,
    'rosy brown': 0xbc8f8f,
    'royal blue 1': 0x4876ff,
    'royal blue 2': 0x436eee,
    'royal blue 3': 0x3a5fcd,
    'royal blue 4': 0x27408b,
    'royal blue': 0x4169e1,
    'saddle brown': 0x8b4513,
    'salmon 1': 0xff8c69,
    'salmon 2': 0xee8262,
    'salmon 3': 0xcd7054,
    'salmon 4': 0x8b4c39,
    'salmon': 0xfa8072,
    'sandy brown': 0xf4a460,
    'sap green': 0x308014,
    'sea green 1': 0x54ff9f,
    'sea green 2': 0x4eee94,
    'sea green 3': 0x43cd80,
    'sea green 4': 0x2e8b57,
    'sea green': 0x2e8b57,
    'seashell 1': 0xfff5ee,
    'seashell 2': 0xeee5de,
    'seashell 3': 0xcdc5bf,
    'seashell 4': 0x8b8682,
    'seashell': 0xfff5ee,
    'sepia': 0x5e2612,
    'sienna 1': 0xff8247,
    'sienna 2': 0xee7942,
    'sienna 3': 0xcd6839,
    'sienna 4': 0x8b4726,
    'sienna': 0xa0522d,
    'silver': 0xc0c0c0,
    'sky blue 1': 0x87ceff,
    'sky blue 2': 0x7ec0ee,
    'sky blue 3': 0x6ca6cd,
    'sky blue 4': 0x4a708b,
    'sky blue': 0x87ceeb,
    'slate blue 1': 0x836fff,
    'slate blue 2': 0x7a67ee,
    'slate blue 3': 0x6959cd,
    'slate blue 4': 0x473c8b,
    'slate blue': 0x6a5acd,
    'slate grey 1': 0xc6e2ff,
    'slate grey 2': 0xb9d3ee,
    'slate grey 3': 0x9fb6cd,
    'slate grey 4': 0x6c7b8b,
    'slate grey': 0x708090,
    'snow 1': 0xfffafa,
    'snow 2': 0xeee9e9,
    'snow 3': 0xcdc9c9,
    'snow 4': 0x8b8989,
    'snow': 0xfffafa,
    'spring green 1': 0x00ee76,
    'spring green 2': 0x00cd66,
    'spring green 3': 0x008b45,
    'spring green': 0x00ff7f,
    'steel blue 1': 0x63b8ff,
    'steel blue 2': 0x5cacee,
    'steel blue 3': 0x4f94cd,
    'steel blue 4': 0x36648b,
    'steel blue': 0x4682b4,
    'tan 1': 0xffa54f,
    'tan 2': 0xee9a49,
    'tan 3': 0xcd853f,
    'tan 4': 0x8b5a2b,
    'tan': 0xd2b48c,
    'teal': 0x008080,
    'thistle 1': 0xffe1ff,
    'thistle 2': 0xeed2ee,
    'thistle 3': 0xcdb5cd,
    'thistle 4': 0x8b7b8b,
    'thistle': 0xd8bfd8,
    'tomato 1': 0xff6347,
    'tomato 2': 0xee5c42,
    'tomato 3': 0xcd4f39,
    'tomato 4': 0x8b3626,
    'tomato': 0xff6347,
    'turquoise 1': 0x00f5ff,
    'turquoise 2': 0x00e5ee,
    'turquoise 3': 0x00c5cd,
    'turquoise 4': 0x00868b,
    'turquoise blue': 0x00c78c,
    'turquoise': 0x40e0d0,
    'violet red 1': 0xff3e96,
    'violet red 2': 0xee3a8c,
    'violet red 3': 0xcd3278,
    'violet red 4': 0x8b2252,
    'violet red': 0xd02090,
    'violet': 0xee82ee,
    'warm grey': 0x808069,
    'wheat 1': 0xffe7ba,
    'wheat 2': 0xeed8ae,
    'wheat 3': 0xcdba96,
    'wheat 4': 0x8b7e66,
    'wheat': 0xf5deb3,
    'white smoke': 0xf5f5f5,
    'white': 0xffffff,
    'yellow 1': 0xffff00,
    'yellow 2': 0xeeee00,
    'yellow 3': 0xcdcd00,
    'yellow 4': 0x8b8b00,
    'yellow green': 0x9acd32,
    'yellow': 0xffff00,
}

"""
Some colors have more than one name.  This is a list of the secondary names for
Juce colors.  We need this to reliably go from color back to color name and
get the same result each time.
"""
SECONDARY_NAMES = {
    'aqua',
    'aquamarine 1',
    'aquamarine 3',
    'azure 1',
    'bisque 1',
    'blue 3',
    'blue 4',
    'chartreuse 1',
    'chocolate 4',
    'corn silk 1',
    'cyan 4',
    'dark olivegreen',
    'deep pink 1',
    'deep sky blue 1',
    'dodger blue 1',
    'fuchsia',
    'gold 1',
    'grey',
    'green 1',
    'honeydew 1',
    'ivory 1',
    'lavender blush 1',
    'lemon chiffon 1',
    'light cyan 1',
    'light salmon 1',
    'light yellow 1',
    'lime',
    'limegreen',
    'magenta 4',
    'medium seagreen',
    'medium slateblue',
    'medium violetred',
    'misty rose 1',
    'navajo white 1',
    'none',
    'olive drab 3',
    'orange 1',
    'orange red 1',
    'pale green 2',
    'peachpuff 1',
    'red 1',
    'red 4',
    'sea green 4',
    'seashell 1',
    'snow 1',
    'tan 3',
    'tomato 1',
    'yellow 1'
}



import ctypes, functools
from . import log

DMX_LENGTH = 512
ARTNET_DMX = 0x5000
ARTNET_VERSION = 14
NAME = b'Art-Net\x00'
MAX_NET = 0xFF
MAX_SUBNET = 0xF
MAX_UNIVERSE = 0xF
UDP_PORT = 0x1936  # 6454


@functools.lru_cache(maxsize=DMX_LENGTH)
def MessageClass(length=DMX_LENGTH):
    assert 0 <= length <= DMX_LENGTH
    assert length % 2 == 0, 'artnet only takes messages of even length'

    Char, Int8, Int16 = ctypes.c_char, ctypes.c_ubyte, ctypes.c_ushort

    class DMXMessage(ctypes.Structure):
        # http://artisticlicence.com/WebSiteMaster/User%20Guides/art-net.pdf p47
        _fields_ = [
            ('id', Char * 8),
            ('opCode', Int16),
            ('protVerHi', Int8),
            ('protVerLo', Int8),
            ('sequence', Int8),
            ('physical', Int8),
            ('subUni', Int8),
            ('net', Int8),
            ('lengthHi', Int8),
            ('length', Int8),
            ('data', Int8 * length),  # At position 18
        ]

    return DMXMessage


def dmx_message(
        length=None, net=0, subnet=0, universe=0, sequence=1, data=None):
    if length is None:
        length = DMX_LENGTH if (data is None) else len(data)
    elif data is not None:
        assert len(data) == length

    Message = MessageClass(length)

    assert 0 <= sequence <= DMX_LENGTH
    assert 0 <= net <= MAX_NET
    assert 0 <= subnet <= MAX_SUBNET
    assert 0 <= universe <= MAX_UNIVERSE

    # http://artisticlicence.com/WebSiteMaster/User%20Guides/art-net.pdf p5
    subUni = (subnet << 4) + universe
    hi, lo = divmod(length, 256)

    msg = Message(
        id=NAME,
        opCode=ARTNET_DMX,
        protVerLo=ARTNET_VERSION,
        sequence=sequence,
        net=net,
        subUni=subUni,
        lengthHi=hi,
        length=lo)

    if data is not None:
        msg.data[:] = data

    return msg


DMXMessage = MessageClass()
EMPTY_MESSAGE_SIZE = ctypes.sizeof(MessageClass(0))


def bytes_to_message(b):
    if not isinstance(b, bytearray):
        b = bytearray(b)

    length = len(b) - EMPTY_MESSAGE_SIZE
    assert 0 <= length <= DMX_LENGTH

    result = MessageClass(length).from_buffer(b)
    if result.id == NAME[:-1]:
        return result
    # For some reason, ctypes throws away that trailing \0 character
    # when converting back to bytes().
    log.error('Expected name %s but got name %s', NAME[:-1], result.id)



import copy, numpy
from . limit import Limit


def is_numpy(x):
    return isinstance(x, numpy.ndarray)


class ListMath:
    @staticmethod
    def clear(color_list):
        if color_list:
            black = tuple(0 for i in color_list[0])
            color_list[:] = (black for i in color_list)

    @staticmethod
    def add(color_list, source, level=1, buffer=None):
        def add(color, src):
            return tuple(int(c + level * s) for (c, s) in zip(color, src))

        if level:
            color_list[:] = (add(c, s) for (c, s) in zip(color_list, source))

        return buffer

    @staticmethod
    def copy(color_list, source):
        color_list[:] = source

    @staticmethod
    def scale(color_list, gain):
        color_list[:] = [tuple(gain * i for i in c) for c in color_list]

    @staticmethod
    def sum(color_list):
        return sum(sum(c) for c in color_list)


class NumpyMath:
    @staticmethod
    def clear(color_list):
        color_list.fill(0)

    @staticmethod
    def add(color_list, source, level=1, buffer=None):
        if level:
            buffer = numpy.multiply(source, level, out=buffer, casting='unsafe')
            color_list += buffer
        return buffer

    @staticmethod
    def copy(color_list, source):
        numpy.copyto(color_list, source, casting='unsafe')

    @staticmethod
    def scale(color_list, gain):
        color_list *= gain

    @staticmethod
    def sum(color_list):
        return sum(sum(c) for c in color_list)


def Math(color_list):
    return NumpyMath if is_numpy(color_list) else ListMath


class Mixer:
    def __init__(self, color_list, sources, levels=None):
        self.color_list = color_list
        self.math = Math(color_list)
        self.sources = sources
        self.levels = list(levels or [])
        needed = len(self.sources) - len(self.levels)
        self.levels.extend(0 for i in range(needed))
        self.buffer = None

    def clear(self):
        self.math.clear(self.color_list)

    def mix(self, master=1):
        for source, level in zip(self.sources, self.levels):
            self.buffer = self.math.add(
                self.color_list, source, level * master, self.buffer)


def to_triplets(colors):
    """
    Coerce a list into a list of triplets.

    If `colors` is a list of lists or strings, return it as is.  Otherwise,
    divide it into tuplets of length three, silently discarding any extra
    elements beyond a multiple of three.
    """
    try:
        colors[0][0]
        return colors
    except:
        pass

    # It's a 1-dimensional list
    extra = len(colors) % 3
    if extra:
        colors = colors[:-extra]
    return list(zip(*[iter(colors)] * 3))



from . rotation import rotate_and_flip
from . matrix import make_matrix_coord_map


def make_cube_coord_map(dx, dy, dz, xy_serpentine=True, offset=0,
                        xy_rotation=0, z_rotation=0,
                        y_flip=False, z_flip=False):
    result = []
    plane_offset = offset
    for z in range(dz):
        plane = make_matrix_coord_map(dx, dy, serpentine=xy_serpentine,
                                      offset=plane_offset, rotation=xy_rotation,
                                      y_flip=y_flip)
        plane_offset += (dx * dy)
        result.append(plane)

    result = rotate_and_flip(result, z_rotation, z_flip)

    return result


def make_cube_coord_map_positions(coord_map):
    num = 0
    dx, dy, dz = (0, 0, len(coord_map))

    for xy in coord_map:
        if len(xy) > dx:
            dy = len(xy)
        for x in xy:
            if len(x) > dx:
                dx = len(x)
            num += len(x)

    result = [None] * num
    for z in range(dz):
        for y in range(dy):
            for x in range(dx):
                result[coord_map[z][y][x]] = [x, y, z]
    return result



from ... util.colors.conversions import color_cmp


def sorter(x):
    if len(x) < 2:
        return

    prev = x[0]
    for i in range(1, len(x)):
        next = x[i]
        if color_cmp(prev, next) > 0:
            x[i - 1], x[i] = next, prev
            return
        prev = next


FUNCTIONS = {'sorter': sorter}
